from enum import Enum
import io
from types import UnionType
from typing import Dict, List, Any, cast, get_args, get_origin
from abc import abstractmethod
import subprocess
import claripy

from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.exceptions import CommandError, DebuggerCommandError, InvalidArgumentError, ValueError, KeyError
from dAngr.utils import AngrValueType
from dAngr.utils.utils import DataType, Operator, check_signature_matches, is_indexable
from contextlib import redirect_stdout, redirect_stderr

from dAngr.utils.loggers import get_logger
log = get_logger(__name__)



class Expression:
    def debugger(self, context):
        from dAngr.cli.command_line_debugger import CommandLineDebugger
        return cast(CommandLineDebugger, context.root.debugger)

    @abstractmethod
    def __call__(self, context:ExecutionContext):
        raise NotImplementedError
    
    @staticmethod
    def toBool(v):
        if v is None:
            return False
        if isinstance(v, bool):
            return v
        elif isinstance(v, str):
            return v != ""
        elif isinstance(v, list):
            return len(v) > 0
        elif not v is None:
            if isinstance(v, int):
                return v > 0
            else:
                return True
        return False


    def __repr__(self):
        # class name followed by __str__
        return f"{self.__class__.__name__}({self.__str__()})"

    @abstractmethod
    def __str__(self):
        raise NotImplementedError    
    @abstractmethod
    def __eq__(self, other):
        raise NotImplementedError
    @abstractmethod
    def __hash__(self):
        return hash(self.__repr__())


class Object(Expression):
    # a symbol is a variable, argument, constant, memory, register, or stream
    def __call__(self, context: ExecutionContext):
        return self.get_value(context)
    @abstractmethod
    def get_value(self, context)-> Any:
        raise NotImplementedError  
    @abstractmethod
    def set_value(self, context, value:Any):
        raise NotImplementedError
    
class Primitive(Object):
    pass

class Literal(Primitive):
    def __init__(self, value):
        self.value = value
        
    def get_value(self, context):
        return self.value
    
    def set_value(self, context, value):
        raise ValueError("Cannot set value to a literal object")
    
    def __str__(self):
        return f"{self.value}"
    
    def __repr__(self):
        return self.__str__()
    
    def __eq__(self, other):
        return isinstance(other,Literal) and self.value == other.value


class Range(Primitive):
    def __init__(self, start:Expression, end:Expression|None=None, step:Expression|None=None):
        self.start = start
        self.end = end
        self.step = step
    def __str__(self):
        if not self.end:
            return f"range({self.start})"
        elif not self.step:
            return f"range({self.start},{self.end})"
        else:
            return f"range({self.start},{self.end},{self.step})"

    def __eq__(self, other):
        return isinstance(other, Range) and self.start == other.start and self.end == other.end and self.step == other.step
    
    def get_value(self, context):
        start = self.start(context) if isinstance(self.start, Expression) else self.start
        if not self.end:
            return range(start)
        end = self.end(context) if isinstance(self.end, Expression) else self.end
        if not self.step:
            return range(start, end)
        step = self.step(context) if isinstance(self.step, Expression) else self.step
        return range(start, end, step)
    
    def set_value(self, context, value):
        raise ValueError("Cannot set value to a range object")
    
class Listing(Primitive):
    def __init__(self, items:List[Object]):
        self.items = items

    def __str__(self):
        return f"{self.items}"
    
    def __eq__(self, other):
        return isinstance(other, Listing) and self.items == other.items

    def get_value(self, context):
        return [i.get_value(context) for i in self.items]
    
    def set_value(self, context, value: list):
        for i, v in enumerate(value):
            self.items[i].set_value(context, v)

class Dictionary(Primitive):
    def __init__(self, items:Dict[str,Object]):
        self.items = items

    def __str__(self):
        return '{' + ", ".join([f"{k}:{v}" for k,v in self.items.items()]) + '}'
    def __eq__(self, other):
        return isinstance(other, Dictionary) and self.items == other.items
    
    def get_value(self, context):
        return {k:v.get_value(context) for k,v in self.items.items()}
    
    def set_value(self, context, value: dict):
        for k, v in value.items():
            self.items[k].set_value(context, v)

#Reference Objects
class ReferenceObject(Object):

    def __init__(self, name:Expression):
        self._name = name

    @staticmethod
    def createNamedObject(db:str, name:Expression):
        switcher = {
            "&reg": Register,
            "&sym": SymbolicValue,
            "&io": Stream,
            "&vars": VariableRef
        }
        return switcher.get(db,None)(name)
        
        
class Stream(ReferenceObject):
    def __init__(self, stream:Expression):
        super().__init__(stream)
    @property
    def stream(self):
        return self._name
    
    def __str__(self):
        return f"&io.{self.stream}"
    def get_value(self, context):
        return self.debugger(context).get_stream(self.stream(context))
    def set_value(self, context, value):
        raise ValueError("Cannot set value to a stream object")
    def __eq__(self, other):
        return isinstance(other, Stream) and self.stream == other.stream

class SymbolicValue(ReferenceObject):
    def __init__(self, name):
        super().__init__(name)

    def get_value(self, context):
        #get variable value and 
        return self.debugger(context).get_symbol(self._name(context))
    
    def set_value(self, context, value):
        assert isinstance(value,AngrValueType)
        self.debugger(context).set_symbol(self._name(context), value)

    def __str__(self):
        return f"&sym.{self._name}"
    def __eq__(self, other):
        return isinstance(other, SymbolicValue) and self._name == other._name
    
class Memory(ReferenceObject):
    def __init__(self, address:Expression, size:Expression|None):
        super().__init__(Literal(f"{address}->{size}"))
        self.address = address
        self.size = size 

    def get_value(self, context):
        return self.debugger(context).get_memory(self.address(context), self.size(context) if self.size else 0)

    def set_value(self, context, value):
        assert isinstance(value,AngrValueType)
        self.debugger(context).set_memory(self.address(context), value, self.size(context) if self.size else None)
    
    def __str__(self):
        if isinstance(self.address, Literal):
            return f"&mem[{hex(self.address.value)}->{self.size}]"
        else:
            return f"&mem[{self.address}->{self.size}]"
    def __eq__(self, other):
        return isinstance(other, Memory) and self.address == other.address and self.size == other.size
    
class Register(ReferenceObject):
    def __init__(self, name):
        self.name = name

    @property
    def register(self):
        return self.name
    def get_value(self, context):
        return self.debugger(context).get_register(self.register(context))
    
    def set_value(self, context, value):
        if isinstance(value, int) or isinstance(value, claripy.ast.BV):
            self.debugger(context).set_register(self.register(context), value)
        else:
            raise ValueError(f"Invalid value type: {value}")    
    def __str__(self):
        return f"&reg.{self.register}"
    def __eq__(self, other):
        return isinstance(other, Register) and self.register == other.register
    
class StateObject(ReferenceObject):
    def __init__(self):
        super().__init__(Literal("state"))
    def get_value(self, context):
        return self.debugger(context).current_state
    def set_value(self, context, value):
        raise ValueError("Cannot set value to a state object")
    def __str__(self):
        return f"&state"
    def __eq__(self, other):
        return isinstance(other, StateObject)
class Negate(Expression):
    def __init__(self, expr:Expression):
        self.expr = expr

    def __call__(self, context:ExecutionContext):
        return -self.expr(context)
    
    def __str__(self):
        return f"-{self.expr}"
    
    def __eq__(self, other):
        return isinstance(other, Negate) and self.expr == other.expr
    
class Invert(Expression):
    def __init__(self, expr:Expression):
        self.expr = expr

    def __call__(self, context:ExecutionContext):
        return not(self.expr(context))
    
    def __str__(self):
        return f"not {self.expr}"
    
    def __eq__(self, other):
        return isinstance(other, Invert) and self.expr == other.expr
    
class VariableRef(ReferenceObject):
    def __init__(self, name:Expression, is_static=False):
        self.name:Expression = name
        self._is_static = is_static
    
    @property
    def is_static(self):
        return self._is_static
    @property
    def static_name(self):
        if not self._is_static:
            raise ValueError(f"Variable {self.name} is not static")
        return f"@{self.name}"
    
    def __str__(self):
        return f"{self.name}"

    def __eq__(self, other):
        return isinstance(other, VariableRef) and self.name == other.name
    
    def var_exists(self, context):
        return self.name(context) in context.variables

    
    def get_value(self, context):
        if self.var_exists(context):
            return context[self.name(context)].value
        sym = self.debugger(context).find_symbol(self.name(context))
        if not sym is None:
            return sym
        raise KeyError(f"Unknown variable: {self.name}")
        
    def set_value(self, context, value):
        context[self.name(context)] = value

class Property(ReferenceObject):
    def __init__(self, obj:Object, prop:str):
        super().__init__(Literal(f"{obj}.{prop}"))
        self.obj = obj
        self.prop = prop

    def __str__(self):
        return f"{self.obj}.{self.prop}"
    def __eq__(self, other):
        return isinstance(other, Property) and self.prop == other.prop
    
    def set_value(self, context, value: Any):
        o = self.obj.get_value(context) 
        if hasattr(o, self.prop):
            setattr(o, self.prop, value)
        else:
            raise InvalidArgumentError(f"Object {o} does not have property {self.prop}")
    def get_value(self, context):
        o = None
        if isinstance(self.obj, VariableRef) and not self.obj.var_exists(context):
            if e := context.find_enum(self.obj.name(context)):
                o = e
            else:
                raise InvalidArgumentError(f"Unknown variable: {self.obj}")
        else:
            o = self.obj.get_value(context)
        if hasattr(o, self.prop):
            return getattr(o, self.prop)
        else:
            raise InvalidArgumentError(f"Object {o} does not have property {self.prop}")
        
class IndexedProperty(Property):
    def __init__(self, obj:Object, index:Expression):
        super().__init__(obj, f"{index}")
        self.index = index
    
    def __str__(self):
        return f"{self.obj}[{self.index}]"
    
    def set_value(self, context, value: Any):
        o = self.obj.get_value(context)
        if is_indexable(o):
            index = self.index(context)
            o[index] = value
        else:
            raise InvalidArgumentError(f"Object {o} does not support indexing")
    def get_value(self, context):
        o = self.obj.get_value(context)
        index = self.index(context)
        if is_indexable(o):
            return o[index]
        else:
            raise InvalidArgumentError(f"Object {o} does not support indexing")

class Slice(Property):
    def __init__(self, obj:Object, start:int|Expression, end:int|Expression):
        super().__init__(obj, f"{start}:{end}")
        self.start = start
        self.end = end
    
    def get_value(self, context):
        start = self.start(context) if isinstance(self.start, Expression) else self.start
        end = self.end(context) if isinstance(self.end, Expression) else self.end
        if isinstance(start,int) and start == -1:
            if isinstance(end,int) and end == -1:
                return self.obj.get_value(context)
            else:
                return self.obj.get_value(context)[:end]
        else:
            if isinstance(end,int) and end == -1:
                return self.obj.get_value(context)[start:]
            else:
                return self.obj.get_value(context)[start:end]
    
    def set_value(self, context, value: list):
        o = self.obj.get_value(context)
        if isinstance(value, list):
            start = self.start(context) if isinstance(self.start, Expression) else self.start
            end = self.end(context) if isinstance(self.end, Expression) else self.end
            o[start:end] = value
        else:
            raise InvalidArgumentError(f"Object {o} does not support slicing")
    def __repr__(self):
        return f"{self.obj}[{self.start}:{self.end}]"
    def __eq__(self, other):
        return isinstance(other, Slice) and self.start == other.start and self.end == other.end
    
class Inclusion(Expression):
    def __init__(self, obj:Object, item:Object, exclude:bool=False):
        self.obj = obj
        self.item = item
        self.exclude = exclude
    
    def __call__(self, context: ExecutionContext):
        o = self.obj.get_value(context)
        i = self.item.get_value(context)
        if self.exclude:
            return o not in i
        return o in i
    
    def __str__(self):
        if self.exclude:
            return f"{self.obj} not in {self.item}"
        return f"{self.item} in {self.obj}"
    
    def __eq__(self, other):
        return isinstance(other, Inclusion) and self.obj == other.obj and self.item == other.item and self.exclude == other.exclude
    
class Comparison(Expression):
    def __init__(self, left:Expression, operator:Operator, right:Expression):
        self.left = left
        self.operator:Operator = operator
        self.right = right
        self.init = False


    def reorder_comparison(self,comp):
        # Base case: If it's not a Comparison, return it as is
        if not isinstance(comp, Comparison):
            return comp
        
        # Recursively reorder left and right comparisons first
        left = self.reorder_comparison(comp.left)
        right = self.reorder_comparison(comp.right)

        # If the right is a Comparison and has higher precedence than current,
        # restructure to respect precedence
        if isinstance(right, Comparison) and right.operator.precedence < comp.operator.precedence: 
            # Re-arrange: the right side will become a new left-side comparison
            new_left = Comparison(left, comp.operator, right.left)
            return Comparison(new_left, right.operator, right.right)
        
        # Otherwise, just return the current comparison
        return Comparison(left, comp.operator, right)

    def __call__(self, context:ExecutionContext):
        # handle operator precedence, the right side of the expression may be a comparison
        if not self.init:
            self = self.reorder_comparison(self)
            self.init = True

        left = self.left(context)
        right = self.right(context)
        switch = {
            Operator.AND: claripy.And,
            Operator.OR: claripy.Or,
        }
        if self.operator == Operator.ADD and isinstance(left, claripy.ast.BV) and isinstance(right, claripy.ast.BV):
            return claripy.Concat(left, right)
        elif isinstance(left, claripy.ast.Base) or isinstance(right, claripy.ast.Base):
            op = switch.get(self.operator, None)
            if op:
                v= op(left, right)
                if v == NotImplemented:
                    raise ValueError(f"Invalid operator {self.operator} for {type(left)} and {type(right)}")
                return v
        operation = Operator.convert_to_magic_method(self.operator)
        if operation is None:
            raise ValueError(f"Invalid operator {self.operator}")
        v = getattr(type(left), operation)(left, right)
        if v == NotImplemented:
            v = getattr(type(right), operation)(left, right)
            if v == NotImplemented:
                raise ValueError(f"Invalid operator {self.operator} for {type(left)} and {type(right)}")
        return v
    
    def __str__(self):
        return f"{self.left} {self.operator} {self.right}"
    
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Comparison) and self.left == value.left and self.operator == value.operator and self.right == value.right
    
class Constraint(Expression):
    pass
class IfConstraint(Constraint):
    def __init__(self, condition:Expression, true_constraint:Constraint, false_constraint:Constraint):
        self.condition = condition
        self.true_constraint = true_constraint
        self.false_constraint = false_constraint

    def __call__(self, context:ExecutionContext):
        cthen =  self.true_constraint(context)
        if not isinstance(cthen, claripy.ast.Base):
            cthen = self.debugger(context).render_argument(cthen,False)
        celse = self.false_constraint(context)
        if not isinstance(celse, claripy.ast.Base):
            celse = self.debugger(context).render_argument(celse,False)
        return claripy.If(self.condition(context), cthen, celse)
    
    def __str__(self):
        return f"if {self.condition} then {self.true_constraint} else {self.false_constraint}"
    
    def __eq__(self, value: object) -> bool:
        return isinstance(value, IfConstraint) and self.condition == value.condition and self.true_constraint == value.true_constraint and self.false_constraint == value.false_constraint
    
class Command(Expression):
    def _merge_consecutive_literals(self, content:List[Literal|Expression]) ->List[Expression]:
        cc = []
        for c in content:
            if cc and isinstance(cc[-1], Literal) and isinstance(c, Literal):
                cc[-1] = Literal(cc[-1].value + str(c.value))
            else:
                cc.append(c)
        return cc
    @abstractmethod
    def __call__(self, context):
        raise NotImplementedError
    
class BASECommand(Command):
    def __init__(self, base:str, args= None):
        self.base = base
        self.args = args
        self.return_value = None

    def __call__(self, context:ExecutionContext):
        if self.args:
            self.return_value = to_val(self.args, context)
        return self
    
    def __str__(self):
        return f"{self.base}"
    def __eq__(self, other):
        return isinstance(other, BASECommand) and self.base == other.base

BREAK = BASECommand('break')
CONTINUE = BASECommand('continue')

def to_val(arg, context:ExecutionContext):
    v = arg(context)
    if isinstance(arg, ReferenceObject):
        if isinstance(v, str) or isinstance(v, bytes):
            return repr(v)
    if isinstance(v, claripy.ast.Base):
        from dAngr.cli.command_line_debugger import dAngrExecutionContext
        dbg = cast(dAngrExecutionContext, context.root).debugger
        v = dbg.eval_symbol(v, DataType.bytes, endness =dbg.project.arch.memory_endness) # type: ignore
        return repr(dbg.cast_to(v, DataType.str))
    return v

class PythonCommand(Command):
    def __init__(self, *cmds:Expression|str|int|bytes, **kwargs:Expression):
        cc = [ Literal(c) if isinstance(c, (str,int,bytes)) else c for c in cmds if isinstance(c, (Expression,str,int,bytes))]
        self.cmds:List[Expression] = self._merge_consecutive_literals(cc)
        self.kwargs:Dict[str,Expression] = kwargs

    
    def __call__(self, context:ExecutionContext):
        # copy the commands locally
        # execute the commands if it is an Expression and replace the entry in the copy with the result
        c = context.clone()
        results = [to_val(a,c) for a in self.cmds] + [f"{k}={to_val(v,c)}" for k,v in self.kwargs.items()]
        stdout_buffer = io.StringIO()
        stderr_buffer = io.StringIO()
        
        with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
            try:
                cmd = "".join([str(r) for r in results])
                vars = context.python_context.get("global",{})
                glo = context.python_context.get("local", {})
                # split the last line, execute the first lines, and eval the last line
                if "\n" in cmd:
                    lines,line = cmd.rsplit("\n",1)
                else:
                    lines, line = "", cmd
                
                exec(lines,glo,vars)
                try:
                    context.return_value = eval(line,glo,vars)
                except SyntaxError:
                    exec(line,glo,vars)
                context.python_context = {"global":glo, "local":vars}
            except Exception as e:
                raise CommandError(f"Error executing python command: {cmd} ({e})")
        if stdout := stdout_buffer.getvalue():
            self.debugger(context).conn.send_output(stdout)
        if stderr := stderr_buffer.getvalue():
            self.debugger(context).conn.send_error(stderr)
        return context.return_value
    
    def __str__(self):
        return "".join([ str(c) for c in self.cmds])
    def __eq__(self, other):
        return isinstance(other, PythonCommand) and self.cmds == other.cmds
    
class BashCommand(Command):
    def __init__(self, *cmds:Expression|str|int|bytes):
        cc = [ Literal(c) if isinstance(c, (str,int,bytes)) else c for c in cmds if isinstance(c, (Expression,str,int,bytes))]
        self.cmds:List[Expression] = self._merge_consecutive_literals(cc)

    def __call__(self, context):
        c = context.clone()
        args = "".join([str(a(c)) for a in self.cmds]).split(" ")
        context.return_value = subprocess.run(args, capture_output=True, text=True, shell=True).stdout.strip()
        return context.return_value
    
    def __str__(self):
        return "".join([ str(c) for c in self.cmds])
    def __eq__(self, other):
        return isinstance(other, BashCommand) and self.cmds == other.cmds
         
class DangrCommand(Command):
    def __init__(self, cmd:str, package:str|None, *args:Expression, **kwargs:Expression):
        self.cmd = cmd
        self.package = package
        self.args:List[Expression] = [*args]
        self.kwargs:Dict[str,Expression] = kwargs

    def _get_enum_value(self, name, spec):
        if get_origin(spec.dtype) is UnionType:
            types = get_args(spec.dtype)
        else:
            types = [spec.dtype]

        if Enum in types:
            for t in types:
                if issubclass(t, Enum):
                    if name in t.__members__: # type: ignore
                        return t[name] # type: ignore
        return None
    def _accept_string(self, name, spec):
        if get_origin(spec.dtype) is UnionType:
            types = get_args(spec.dtype)
        else:
            types = [spec.dtype]
        if str in types:
            return name
        return None

    def _check_arg(self, arg, context:ExecutionContext, spec):

        #TODO, what if  dtype is a tuple
        # if spec.dtype == str and isinstance(arg, VariableRef):
        #     return arg.name
        if isinstance(spec.dtype, type) and issubclass(spec.dtype, ReferenceObject):
            return arg
        elif isinstance(spec.dtype, type) and issubclass(spec.dtype, Enum):
            if not isinstance(arg.name, Literal):
                return spec.dtype[arg.value]
            else:
                return spec.dtype[arg.name(context)]
        else:
            if isinstance(arg, VariableRef):
                if arg.var_exists(context):
                    return arg(context)
                name = arg.name(context)
                if e:= self._get_enum_value(name, spec):
                    return e
                if s:= self._accept_string(name, spec):
                    return s
                raise DebuggerCommandError(f"Unknown variable: {arg.name}")
        return arg(context) 

        

    def __call__(self, context:ExecutionContext):
        from dAngr.cli.command_line_debugger import BuiltinFunctionDefinition
        spec = None
        spec = context.find_function(self.package, self.cmd)
        if not spec:
            if self.package:
                raise CommandError(f"Unknown command: {self.package}.{self.cmd}")
            # if context.find_variable(self.cmd):
            #     # make an expression with the var and args
            #     #TODO: what if not comparison?
            #     if self.args:
            #         c = Comparison(VariableRef(Literal(self.cmd)), Operator.ADD, self.args[0])
            #         return c(context)
            #     else:
            #         return context[self.cmd].value
            # elif e:= context.find_enum(self.cmd):
            #     return e
            else:
                raise CommandError(f"Unknown command: {self.cmd}")
        # spec = cast(BuiltinFunctionDefinition, spec)
        #handle function arguments
        s_args = spec.args

        if len(spec.args) < len(self.args):
            #handle variable arguments
            if spec.args and  spec.args[-1].dtype == tuple:
                # copy the last argument until there are enough arguments
                while len(s_args) < len(self.args):
                    s_args.append(s_args[-1])
            else:
                raise CommandError(f"Too many arguments. Expected {len(spec.args)} but got {len(self.args)}")
        # check arg type and if the arguments match the function signature
        arguments = [ self._check_arg(arg, context, s_args[i]) for i,arg in enumerate(self.args) ]
        named_args = {k: self._check_arg(v,context, spec.get_arg_by_name(k)) for k,v in self.kwargs.items()}
        if isinstance(spec, BuiltinFunctionDefinition):
            # check if the function signature matches
            check_signature_matches(spec.func, spec, arguments, named_args)

        log.debug(lambda:f"Calling {self.cmd} with {arguments} {named_args}")

        # call the function
        context.return_value = spec(context, *arguments, **named_args)
        return context.return_value

            

    def __str__(self):
        return f"{self.cmd} {[str(a) for a in self.args]} {self.kwargs}"
    def __eq__(self, other):
        return isinstance(other, DangrCommand) and self.cmd == other.cmd and self.package == other.package and self.args == other.args and self.kwargs == other.kwargs

    


# Generated from /workspaces/dAngr/src/dAngr/cli/grammar/ranges.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .rangesParser import rangesParser
else:
    from rangesParser import rangesParser

# This class defines a complete listener for a parse tree produced by rangesParser.
class rangesListener(ParseTreeListener):

    # Enter a parse tree produced by rangesParser#expression.
    def enterExpression(self, ctx:rangesParser.ExpressionContext):
        pass

    # Exit a parse tree produced by rangesParser#expression.
    def exitExpression(self, ctx:rangesParser.ExpressionContext):
        pass


    # Enter a parse tree produced by rangesParser#range.
    def enterRange(self, ctx:rangesParser.RangeContext):
        pass

    # Exit a parse tree produced by rangesParser#range.
    def exitRange(self, ctx:rangesParser.RangeContext):
        pass


    # Enter a parse tree produced by rangesParser#dangr_range.
    def enterDangr_range(self, ctx:rangesParser.Dangr_rangeContext):
        pass

    # Exit a parse tree produced by rangesParser#dangr_range.
    def exitDangr_range(self, ctx:rangesParser.Dangr_rangeContext):
        pass


    # Enter a parse tree produced by rangesParser#bash_range.
    def enterBash_range(self, ctx:rangesParser.Bash_rangeContext):
        pass

    # Exit a parse tree produced by rangesParser#bash_range.
    def exitBash_range(self, ctx:rangesParser.Bash_rangeContext):
        pass


    # Enter a parse tree produced by rangesParser#python_range.
    def enterPython_range(self, ctx:rangesParser.Python_rangeContext):
        pass

    # Exit a parse tree produced by rangesParser#python_range.
    def exitPython_range(self, ctx:rangesParser.Python_rangeContext):
        pass


    # Enter a parse tree produced by rangesParser#bash_content.
    def enterBash_content(self, ctx:rangesParser.Bash_contentContext):
        pass

    # Exit a parse tree produced by rangesParser#bash_content.
    def exitBash_content(self, ctx:rangesParser.Bash_contentContext):
        pass


    # Enter a parse tree produced by rangesParser#py_content.
    def enterPy_content(self, ctx:rangesParser.Py_contentContext):
        pass

    # Exit a parse tree produced by rangesParser#py_content.
    def exitPy_content(self, ctx:rangesParser.Py_contentContext):
        pass


    # Enter a parse tree produced by rangesParser#anything.
    def enterAnything(self, ctx:rangesParser.AnythingContext):
        pass

    # Exit a parse tree produced by rangesParser#anything.
    def exitAnything(self, ctx:rangesParser.AnythingContext):
        pass


    # Enter a parse tree produced by rangesParser#symbol.
    def enterSymbol(self, ctx:rangesParser.SymbolContext):
        pass

    # Exit a parse tree produced by rangesParser#symbol.
    def exitSymbol(self, ctx:rangesParser.SymbolContext):
        pass



del rangesParser


from abc import abstractmethod
import inspect
from typing import Any, Callable, cast, get_args
from angr import SimulationManager
from typing import List

from dAngr.angr_ext.debugger import Debugger
from dAngr.angr_ext.step_handler import StopReason
from dAngr.cli.grammar.definitions import ArgumentSpec, FunctionDefinition
from dAngr.cli.grammar.execution_context import Variable
from dAngr.exceptions import ExecutionError, InvalidArgumentError,DebuggerCommandError
from dAngr.utils.utils import str_to_type, undefined, AngrType, AngrValueType, SymBitVector

# required for str_to_type - do not remove
from dAngr.cli.grammar.expressions import *




# def get_cmd_name(cls):
#     return ''.join(['_'+i.lower() if i.isupper() else i for i in cls.__name__.replace('Command', '')]).lstrip('_')

def get_short_cmd_name(name):
    # use the first letter of each word in the command name
    return ''.join([i.lower() for i in name if i.isupper()])

class AutoRunMeta(type):
    def __init__(self, name, bases, dct):
        super().__init__(name, bases, dct)
        # Automatically trigger the static method
        if hasattr(self, '__render_commands__'):
            self.__render_commands__(self) # type: ignore

class IBaseCommand(metaclass=AutoRunMeta):
    def __init__(self, debugger:Debugger):
        self._debugger = debugger
    
    def cmd_specs(self):
        return getattr(self.__class__, "__cmd_specs__", {})
    
    def get_cmd_specs(self, command:str):
        specs = self.cmd_specs()
        return next((specs[s] for s in specs if specs[s].name.strip() == command or specs[s].short_name == command), None)
    
    @abstractmethod
    def execute(self, cmd, *args):
        pass



class BuiltinFunctionDefinition(FunctionDefinition):
    def __init__(self, name:str, cmd_class:type[IBaseCommand], func: Callable,  description:str, args:List[ArgumentSpec], short_name:str, example:str, package:str):
        super().__init__(name, package, args)
        self._cmd_class = cmd_class
        self._func = func
        self._description = description
        self._short_name = short_name
        self._example = example
        self._package = package
    @property
    def name(self):
        return self._name
    @property
    def func(self):
        return self._func
    @property
    def cmd_class(self):
        return self._cmd_class
    @property
    def description(self):
        return self._description
    @property
    def short_name(self):
        return self._short_name
    @property
    def example(self):
        return self._example
    @property
    def package(self):
        return self._package
    
    def __call__(self, context:ExecutionContext,*args, **named_args) -> Any:
        from dAngr.cli.command_line_debugger import dAngrExecutionContext
        o = self._cmd_class(cast(dAngrExecutionContext,context.root).debugger)
        # get the function from the class
        f = getattr(o, self._name, None)
        if not f:
            f = getattr(o, self._name + '_')

        return f(*args, **named_args)
    
def convert_args(args, signature):
    # convert the args to the correct type
    pargs = []
    for name in args:
        a = signature.parameters.get(name)
        if not a:
            raise InvalidArgumentError(f"Function {signature} does not have a parameter named {name}")
        arg = args[name]
        tp = arg["type"]
        #check if the type matches the function definitions arg type
        if a.kind == inspect.Parameter.VAR_POSITIONAL:
            if tp != tuple:
                raise InvalidArgumentError(f"Function {signature} parameter {name} has type {a.annotation} but expected {tp}")
            else:
                # get the type of the elements in the tuple
                pargs.append(ArgumentSpec(name, tp, undefined, arg["description"]))
        elif a.kind == inspect.Parameter.VAR_KEYWORD:
            if tp != dict:
                raise InvalidArgumentError(f"Function {signature} parameter {name} has type {a.annotation} but expected {tp}")
            else:
                # get the type of the elements in the tuple
                pargs.append(ArgumentSpec(name, tp, undefined, arg["description"]))
        else:
            if tp is None:
                tp = inspect._empty
                # raise InvalidArgumentError(f"Function {signature} parameter {name} does not have a type annotation")
            if tp != a.annotation:
                # check if the type is a union
                if not (tp in get_args(a.annotation)):
                    raise InvalidArgumentError(f"Function {signature} parameter {name} has type {a.annotation} but expected {tp}")
            description = arg["description"]
            default = a.default if a.default != inspect._empty else undefined
            pargs.append(ArgumentSpec(name,tp, default, description))  # type: ignore
    return pargs


def parse_docstring(docstring:str):
    # parse the description, args, short name, and extra info, return value from the docstring, Don't care about the errors raised
    description = ""
    args = []
    example = ""
    short_name = ""
    return_value = ""

    state = 0
    if docstring:
        lines = docstring.split("\n")
        description = lines[0]
        for line in lines[1:]:
            line = line.strip()
            if not line:
                continue
            if line.lower().startswith("args:"):
                state = 1
                if line.endswith(":"):
                    continue
            elif line.lower().startswith("short name:"):
                if line.endswith(":"):
                    continue
                state = 2
            elif line.lower().startswith("returns:"):
                state = 3
                if line.endswith(":"):
                    continue
            elif line.lower().startswith("raises:"):
                state = 4
                if line.endswith(":"):
                    continue
            elif line.lower().startswith("example:"):
                state = 5
                if line.endswith(":"):
                    continue
            if state == 0:
                description += line
            elif state == 1:
                args.append(line)
            elif state == 2:
                short_name = line.split(":")[1].strip()
            elif state == 3:
                return_value = line
            elif state == 4:
                # parse the raises
                pass
            if state == 5:
                example += line
    args = parse_args(args)

    return {"description":description, "args":args, "short_name":short_name, "example":example, "return_value":return_value}



def parse_args(args):
    # parse the args
    parsed_args = {}
    for arg in args:
        arg = arg.strip()
        if arg:
            if ":" in arg:
                name_type, description = arg.split(":", 1)
                name, dtype = name_type.strip().split(' ')
                dtype = dtype.strip('(').strip(')').strip()
                #convert string dtype to typings type
                tp = str_to_type(dtype)
                parsed_args[name] = {"type":tp, "description":description.strip()}
            else:
                # TODO add multiline support
                raise InvalidArgumentError(f"Invalid argument specification: {arg}")  
    return parsed_args


class BaseCommand(IBaseCommand, metaclass=AutoRunMeta):
    @staticmethod
    def __render_commands__(base_class):
        # through reflection get each function from the class
        # if the function is public, add it as a command
        if not issubclass(base_class, __class__) or base_class == __class__:
            return

        def check(func):
            if inspect.isfunction(func):
                if func.__name__.startswith('_') or func.__name__ == 'cmd_specs':
                    return False
                # check if the function is from the derived class
                if func.__qualname__.split('.')[0] == base_class.__name__:
                    return True
            return False
        functions = inspect.getmembers(base_class, predicate=check)
        specs = []
        for name, fun in functions:
            name = name.strip('_')
            if name == 'execute':
                raise DebuggerCommandError("Function name 'execute' is reserved.")
            doc = fun.__doc__
            if not doc:
                raise DebuggerCommandError(f"Function {name} does not have a docstring")
            #parse the docstring
            info = parse_docstring(doc)
            args = convert_args(info["args"], signature=inspect.signature(fun))
            package = base_class.__module__.split('.')[-1]
            #parse the arguments
            specs.append(BuiltinFunctionDefinition(name,base_class, fun, info["description"], args, info["short_name"], info["example"], package))
        setattr(base_class, "__cmd_specs__", {s.name: s for s in specs})
        # args = []
        # oargs = []
        # named_args = decorator_kwargs
        # # check if args are optional
        # signature = inspect.signature(fun)
        # for arg in named_args:
        #     if arg == "short_name":
        #         continue
        #     par_description = named_args[arg]
        #     if arg not in signature.parameters:
        #         raise ValueError(f"Function {fun.__name__} does not have a parameter named {arg}")
        #     t = signature.parameters[arg].annotation
        #     d = signature.parameters[arg].default
        #     if not d is inspect._empty:
        #         oargs.append((arg,t,par_description,d))
        #     else:
        #         if oargs:
        #             raise ValueError("Optional arguments must come after required arguments")
        #         args.append((arg,t,par_description))
        # name = fun.__name__.strip('_')
        # short_cmd_name = named_args.get("short_name", get_short_cmd_name(name))
        # BaseCommand.cmd_specs.append({"name": name, "func":fun, "description": description, "args":args, "optional":oargs, short_cmd_name: short_cmd_name, "extra_info": extra_info})


    def __init__(self, debugger:Debugger):
        super().__init__(debugger)
        self.paused = False


    @property
    def debugger(self):
        from dAngr.cli.command_line_debugger import CommandLineDebugger

        if self._debugger is None:
            raise ExecutionError("Debugger not set.")
        # return self._debugger
        return cast(CommandLineDebugger,self._debugger)
    
    def to_value(self, value: int|bytes|str|SymBitVector|Variable):
        if value is None:
            return None
        if isinstance(value, Variable):
            return value.value
        return value
    
    def get_angr_value(self, ref:AngrType)->AngrValueType:
        if isinstance(ref, str):
            var = self.debugger.context.find_variable(ref)
            if var:
                assert isinstance(var.value, AngrValueType), f"Invalid value type {type(var.value)}"
                return var.value
            else:
                sym = self.debugger.find_symbol(ref)
                if not sym is None:
                    return sym
                else:
                    log.debug(f"Variable or symbol {ref} not found, asuming string conversion.")
                    return ref
        elif isinstance(ref, Variable):
            return ref.value # type: ignore
        else:
            return ref

    def run_angr(self, until:Callable[[SimulationManager],StopReason] = lambda _: StopReason.NONE, single_step = False):
        u = until
        self.debugger.run(u, single_step=single_step)

    # def get_example(self):
    #     args_lst = [f"<{a[0].replace(' ','_')}>"  for a in self.arg_specs]
    #     options = [f"<{a[0].replace(' ','_')}>"  for a in self.optional_args]
    #     args = ''
    #     if args_lst:
    #         args = ', '.join(args_lst)
    #     if args and options:
    #         args += ', ['
    #         args += ', '.join(options)
    #         args += ']'
    #     if args:
    #         args = " " + args  
    #         return f"{get_cmd_name(self.__class__)}{args}"
    #     else:
    #         return None
    
    def send_info(self, data):
        return self.debugger.conn.send_info(data)

    def send_error(self, data):
        return self.debugger.conn.send_error(data)
    
    def send_warning(self, data):
        return self.debugger.conn.send_warning(data)
    
    def send_result(self, data, newline = True, style = None):
        return self.debugger.conn.send_result(data, newline, style=style)

    

    


from .base import BaseCommand, IBaseCommand
from .breakpoints import BreakpointCommands
from .filters import FilterCommands
from .memory import MemoryCommands
from .execution import ExecutionCommands
from .functions import FunctionCommands
from .info import InformationCommands
from .viewer import ViewerCommands
from .symbols import SymbolCommands
from .files import FileCommands
from .tools import ToolCommands
from .general import GeneralCommands
from .exclusions import ExclusionCommands


from typing import List, Tuple
from prompt_toolkit import HTML, ANSI, print_formatted_text
import html

from dAngr.angr_ext.connection import Connection
from prompt_toolkit.styles import Style

from dAngr.cli import pretty_print

class CliConnection(Connection):

    def __init__(self):
        super().__init__()
        self.indent = 4
        self._history:List[List[Tuple[str,Style|None]]] = [[]]
        self._first = True

    @property
    def output(self)->List[Tuple[str,Style|None]]:
        return self._history[-1]
    
    @output.setter
    def output(self, value:List[Tuple[str,Style|None]]):
        self._history[-1] = value
    
    @property
    def history(self)->List[List[Tuple[str,Style|None]]]:
        return self._history
    
    def clear_history(self):
        self._history = [[]]
        self._first = True

    def clear_output(self):
        if len(self._history[-1])!=0:
            self._history.append([])
        if len(self._history)>5:
            #drop the first element
            self._history.pop(0)
        self._first = True
    @property
    def last_command(self)->str:
        return self._history[-1][0][0] if self._history[-1] else ""

    def _indent(self, data):
         return " "*self.indent + str(data).replace("\n", "\n" + " "*self.indent).replace("\t", " "*self.indent)



    def _escape(self, str_data, esscape_html=True):
        # indent data
        # replace non-printable characters with their hex representation
        data = "".join([c if c.isprintable() or c in ['\r','\n', '\t'] else f"\\x{ord(c):02x}" for c in str_data])
        return html.escape(data ) if esscape_html else str_data
    
    def send_result(self, data, newline:bool=True, style=None):
        if data is None:
            return
        elif isinstance(data, HTML):
            print_formatted_text(data, style=style)
            return
        elif isinstance(data, ANSI):
            print_formatted_text(data, style=style)
            return
        if newline:
            conc = "\n"
        else: conc = ""
        data = pretty_print.pretty_print(data)
        # if data is a list, print each element per line
        data = conc.join([str(x) for x in data]) if isinstance(data, list) else data
        #if data is a dictionary, print each key value pair per line
        data = conc.join([f"{k}: {v}" for k,v in data.items()]) if isinstance(data, dict) else data
        #if data is not a str  by now, convert it to a string
        data = str(data) if not isinstance(data, str) else data

        _data = data
        if newline:
            data = self._indent(data)
        #replace newlines with newlines and 4 spaces
        data = self._escape(data, style is None)
        #get total length of entries in self._output
        l = sum([len(x) for x in self.output]+[len(data)])

        if not style and l > 1000:
            if self._first:
                self._first = False
                print_formatted_text(HTML(self._escape(_data[:1000])),style=style, end = conc)
                print_formatted_text(HTML("<gray>Output too long, type 'less' to view</gray>"),style=style)
        else:
            print_formatted_text(HTML(data),style=style, end=conc)
        self.output.append((data.lstrip(),style))

    def send_output(self, data, style=None):
        #replace newlines with newlines and 4 spaces
        data = pretty_print.pretty_print(data)
        data = self._escape(data)
        print_formatted_text(HTML(f"<skyblue>    > {data}</skyblue>"),style=style)

    def send_info(self, data, style=None):
        #replace newlines with newlines and 4 spaces
        data = pretty_print.pretty_print(data)
        data = self._escape(data)
        print_formatted_text(HTML(f"<green>    Info: {data}</green>"),style=style)

    def send_warning(self, data, style=None):
        #replace newlines with newlines and 4 spaces
        data = pretty_print.pretty_print(data)
        data = self._escape(data)
        print_formatted_text(HTML(f"<yellow>    Warning: {data}</yellow>"),style=style)

    def send_error(self, data, style=None):
        #replace newlines with newlines and 4 spaces
        if isinstance(data, Exception):
            data = str(data)
        data = pretty_print.pretty_print(data)
        data = self._escape(data)
        print_formatted_text(HTML(f"<red>    {data}</red>"),style=style)



import re

import angr
from dAngr.angr_ext.utils import convert_string
from dAngr.cli.debugger_commands import BaseCommand
from dAngr.exceptions.DebuggerCommandError import DebuggerCommandError
from dAngr.exceptions.InvalidArgumentError import InvalidArgumentError
from dAngr.utils.utils import parse_arguments


class FunctionCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)
        
    def get_return_value(self):
        """
        Get the return value after running the function.

        Returns:
            str: The return value of the function.

        Short name: fgr
        """
        val = self.debugger.get_return_value() 
        return str(val) if not val is None else ""

    def set_call_state(self, function_addr: int, *args):
        """
        Set the call state of a function.

        Args:
            function_addr (int): The address of the function.
            args (tuple): Arguments to the function.

        Raises:
            DebuggerCommandError: If the function address is not found.
        
        Short name: scs
        """
        self.debugger.set_call_state(function_addr, args)
        self.send_info(f"Call state set for function at {hex(function_addr)}")

    def set_function_call(self, function_call: str):
        """
        Set the prototype of a function.

        Args:
            function_call (str): C-style function call with arguments.

        Raises:
            DebuggerCommandError: If the function signature format is invalid or if the function address is not found.
        
        Short name: fsc
        """
        # TODO: clean this up
        pattern = r'(\w+)\((.*?)\)'
        match = re.match(pattern, function_call.strip())
        
        if not match:
            raise DebuggerCommandError("Invalid function signature format.")
        
        function_name, vals_str = match.groups()
        func = self.debugger.get_stored_function(function_name)

        if func and "prototype" in func:
            prototype = func["prototype"]
            addr = self.debugger.get_function_address(function_name)
            if addr is None:
                raise DebuggerCommandError(f"Function address not found for {function_name}")
            cc = self.debugger.get_function_cc()
            self.debugger.store_function(function_name, prototype, addr, cc)
        else:
            raise DebuggerCommandError("Prototype not properly initialized, use SetFunctionPrototype command first")
        try:
            arg_strs = parse_arguments(vals_str, ",")
            # base_state = self.debugger.get_current_state()
            arguments = []
            ix = 0
            for value in arg_strs:
                value = value.strip()
                tp = prototype.args[ix]
                v = convert_string(tp, value)
                ix = ix + 1
                if type(tp) is angr.types.SimTypePointer and not type(v) is int:
                    v = 0x1000 * ix
                arguments.append(v)

            state = self.debugger.get_function_callstate(function_name, prototype, cc, arguments)
            self.debugger.set_current_function(function_name)
            ix = 0
            info = []
            for value in vals_str.split(','):
                value = value.strip()
                tp = prototype.args[ix]
                v = convert_string(tp, value)
                ix = ix + 1
                if type(tp) is angr.types.SimTypePointer and not type(v) is int:
                    self.debugger.set_memory(0x1000 * ix, v, len(v)) # type: ignore
                    v = 0x1000 * ix
                    info.append(f"Value {value} stored at {hex(v)}")
            
        except Exception as e:
            raise DebuggerCommandError(f"Error setting up function call: {e}")

        self.send_info(f"Function setup at {hex(addr)} with memory:{info}")
    
    def set_function_prototype(self, prt:str):
        """
        Set the prototype of a function.

        Args:
            prt (str): Prototype of the function.

        Raises:
            InvalidArgumentError: If the function signature format is invalid.
        
        Short name: fsp
        """
        # Define regex pattern for parsing function signature
        pattern = r'(.*)\s+(\w+)\s*\((.*?)\)'
        match = re.match(pattern, prt.strip())
        if not match:
            raise InvalidArgumentError("Invalid function signature format.")

        return_type, function_name, args_str = match.groups()
        self.debugger.set_function_prototype(return_type, function_name, args_str.split(','))

        self.send_info(f"Function signature set for {function_name}")



    def decompiled_function(self, function:str):  # type: ignore
        """
        Show the decompiled function.

        Args:
            function (str): The name of the function.

        Returns:
            str: The decompiled function.

        Raises:
            DebuggerCommandError: If no function is found with the given name.
        
        Short name: fd
        """
        b = self.debugger.get_decompiled_function(function)
        if b is None:
            raise DebuggerCommandError("No basic block found.")
        return f"{b}"
    
    

    def decompiled_function_at_address(self, address:int, end:int):  # type: ignore
        """
        Show the decompiled function at a given address.

        Args:
            address (int): The address to decompile the function at.
            end (int): The end address of the function.

        Returns:
            str: The decompiled function.

        Raises:
            DebuggerCommandError: If no function or basic block is found at the given address.
        Short name: fda
        """
        # func = address
        # func = self.debugger.get_function_info(address)
        # if func is None:
        #     raise DebuggerCommandError("No function found at this address.")
        b = self.debugger.get_decompiled_function_at_address(address, end)
        if b is None:
            raise DebuggerCommandError("No basic block found.")
        return f"{b}"

    def get_function_info(self, function: str|int):
        """
        Get information about a function.

        Args:
            function (str|int): The name or address of the function.

        Returns:
            str: Information about the function.

        Raises:
            DebuggerCommandError: If no function is found with the given name.
        
        Short name: fi
        """
        func = self.debugger.get_function_info(function)

        if not func:
            raise DebuggerCommandError("No function found with this name.")
        
        return func
    



from .CommandError import CommandError
from .InvalidArgumentError import InvalidArgumentError
from .ExecutionError import ExecutionError
from .DebuggerCommandError import DebuggerCommandError
from .ParseError import ParseError
from .ValueError import ValueError
from .FileNotFoundError import FileNotFoundError
from .KeyError import KeyError


from abc import abstractmethod
from enum import Enum, auto

from angr import SimState

class StopReason(Enum):
    STEP = auto()
    BRANCH = auto()
    #PAUSE = auto()
    BREAKPOINT = auto()
    TERMINATE = auto()
    NONE = auto()

class StepHandler:
    
    @abstractmethod
    def handle_output(self, output:str):
        pass

    @abstractmethod
    def handle_step(self, reason:StopReason, state:SimState|None):
        pass
    


import os
import pytest

from unittest.mock import Mock

from dAngr.cli.cli_connection import CliConnection
from dAngr.cli.command_line_debugger import CommandLineDebugger, dAngrExecutionContext
from dAngr.cli.grammar.parser import parse_input
from dAngr.cli.grammar.script import Script


class TestGenericCommandExecution:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)


    
    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_output = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        return dbg

    def test_python_command(self,capsys,dbg):
        input = "!print('x')"
        result = parse_input(input, dbg)
        
        assert isinstance(result, Script)
        ctx = dAngrExecutionContext(dbg,{})
        r = result(ctx)
        assert r == None
        assert str(dbg.conn.send_output.call_args[0][0])  == "x\n"
    
    def test_test_claripy(self,capsys,dbg):
        r = dbg.handle("load 'example'")
        assert r == True
        r = dbg.handle("set_blank_state")
        r = dbg.handle("add_symbol a 1")
        r = dbg.handle("add_symbol b 1")
        r = dbg.handle("add_constraint &sym.a + &sym.b == b'ab'")
        assert r == True, "add_constraints failed"




import inspect
import os

from prompt_toolkit import ANSI
from dAngr.angr_ext.debugger import Debugger
from dAngr.angr_ext.step_handler import StopReason
from dAngr.cli.debugger_commands import BaseCommand
from dAngr.cli.grammar.definitions import FunctionDefinition
from dAngr.cli.script_processor import ScriptProcessor
from dAngr.exceptions import DebuggerCommandError
from dAngr.utils import AngrType
import angr

from dAngr.utils.loggers import AsyncLogger

log = AsyncLogger("execution")

class ExecutionCommands(BaseCommand):
    
    def __init__(self, debugger:Debugger):
        super().__init__(debugger)
    
    # def continue_(self): # type: ignore
    #     """
    #     Run until a breakpoint or terminated. Same as run.

    #     Short name: c
    #     """
    #     super().run_angr()
    
    def run(self): # type: ignore
        """
        Run until a breakpoint or terminated. Same as continue.

        Short name: c
        """
        super().run_angr()

    def step_out(self):  # type: ignore
        """
        Step out of the current function.

        Short name: sO
        """
        cs0 = self.debugger.get_callstack(self.debugger.current_state)
        def check_call_stack(simgr)->StopReason:
            cs = self.debugger.get_callstack(simgr.one_active)
            if len(cs) < len(cs0):
                return StopReason.STEP
            return StopReason.NONE

        super().run_angr(check_call_stack) # return immediately

    def step_over(self): # type: ignore        
        """
        Step over the current statement.

        Short name: so
        """
        cs0 = self.debugger.get_callstack(self.debugger.current_state)
        def check_call_stack(simgr)->StopReason:
            cs = self.debugger.get_callstack(simgr.one_active)
            if len(cs)!= len(cs0):
                return StopReason.NONE
            for i in range(0,len(cs)):
                if cs[i]['func']!= cs0[i]['func']:
                    return StopReason.NONE
            return StopReason.STEP

        self.run_angr(check_call_stack) # return immediately

    def step(self): # type: ignore
        """
        Take a next debugging step (per basic block).

        Short name: s
        """
        super().run_angr(lambda _: StopReason.STEP) # return immediately
        if self.debugger.verbose_step:
            pstr_state = self.debugger.visulize_state()
            self.send_result(ANSI(pstr_state))
        # Throws errors in example 14 of malware analysis
        #return self.debugger.get_current_basic_block() 
    
    def single_step(self):
        """
        Take a next debugging step (per instruction).

        Short name: ss
        """
        super().run_angr(lambda _: StopReason.STEP, single_step=True)
        if self.debugger.verbose_step:
            pstr_state = self.debugger.visulize_state()
            self.send_result(ANSI(pstr_state))

    def get_callable_function(self, addr: int, *args):
        """
        Get a callable function.

        Args:
            addr (int): The address of the function to get.
            args (tuple): The arguments to pass to the function.

        Short name: gcf
        """
        function = self.debugger.get_callable_function(addr)
        return function(*args) 
    
    def exit(self):
        """
        Exit the debugger.

        Short name: q
        """
        self.debugger.stop()

    def load_hooks(self, filename:str):
        """
        Load hooks from a file.

        Args:
            filename (str): The python file containing the hooks to load.

        Example:
            class printf(angr.SimProcedure): 
                def run(self, args):
                    print(f"Running hooked print function in example_hooks.py: {args}")
                    return
        
        Short name: hl
        """
        self.debugger.load_hooks(filename)
        self.send_info(f"Hooks '{filename}' successfully attached.")


   
    
    def _generate_simProcedure(self, func:FunctionDefinition):
        from dAngr.cli.command_line_debugger import CommandLineDebugger
        def generate_function(func):
            def run(self, *args):
                self._run(*args)
            params = [inspect.Parameter(a.name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for a in func.args]
            signature = inspect.Signature([inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)] +params)
            run.__signature__ = signature
            return run
        class CustomSimProcedure(angr.SimProcedure):
            def __init__(self, debugger, func):
                self.debugger = debugger
                self.func:FunctionDefinition = func
                super().__init__()
            def _run(self, *args):
                prev = self.debugger.current_state
                try:
                    self.debugger.current_state = self.state
                    return func(self.debugger.context, *args)
                finally:
                    self.debugger.current_state = prev

        cls = type('CustomSimProcedure_'+func.name, (CustomSimProcedure,), {
            'run': generate_function(func)
        })
        return cls(self.debugger, func)

    def _get_standard_lib(self, func):
        if "." in func:
                lib, func = func.split(".")
        else: lib= 'libc'
        return angr.SIM_PROCEDURES[lib][func]()

    def hook_function(self,  definition:str, target:int|str):
        """
        Add a hook to replace complete function/symbol.

        Args:
            definition (str): The definition of the hook. Standard sim procedures can be addressed as [lib].[func] e.g. libc.printf
            target (int|str): The target to add the hook (int for an address, string for a function name).

        Short name: hf
        """
        func = self.debugger.context.find_definition(definition)
        if func:
            if not isinstance(func, FunctionDefinition):
                raise DebuggerCommandError(f"Definition '{definition}' is not a function definition.")
            proc = self._generate_simProcedure(func)
        else:
            proc = self._get_standard_lib(definition)
        self.debugger.add_function_hook(target, proc)
        self.send_info(f"Hook added at {target}.")

    def hook_region(self,  definition:str, location:int|str, skip_length:int = 0, replace:bool = True):
        """
        Add a hook at a specific location/call site, and skip 'skip_length' instructions.

        Args:
            definition (str): The definition of the hook.
            location (int|str): The location to add the hook (int for an address, string for a function name).
            skip_length (int): The length of the instruction to skip. Default is 0.
            replace (bool): replaces the region. Default = True

        Short name: hr
        """
        if isinstance(location, str):
            address = self.debugger.get_function_address(location)
            if not address:
                raise DebuggerCommandError(f"Function '{location}' not found.")
        else:
            address = location
        func = self.debugger.context.find_definition(definition)
        if not func:
            run_sync = self._get_standard_lib(definition)
        else:
            def run_sync(state,*args):
                try:
                    prev = self.debugger.current_state
                    self.debugger._set_current_state(state)
                    return func(self.debugger.context, *args)
                finally:
                    self.debugger._set_current_state(prev)
        
        self.debugger.add_hook(address, run_sync , skip_length, replace)
        self.send_info(f"Hook added at {location}.")

    def add_to_state(self, name:str, value:AngrType):
        """
        Add a value to the current state.

        Args:
            name (str): The name of the value to add.
            value (AngrType): Value to add to the state.

        Short name: sta
        """
        self.debugger.add_to_state(name, value)
        self.send_info(f"Value {value} added to the state.")

    def get_from_state(self, name:str):
        """
        Get a value from the current state.

        Args:
            name (str): The name of the value to get.

        Short name: sg
        """
        return self.debugger.get_from_state(name)

    def load(self, binary_path:str, base_addr:int=0, veritesting:bool=False, **kwargs):
        """
        Load a binary into the debugger.

        Args:
            binary_path (str): The path to the binary to load.
            base_addr (int): The base address of the binary. Default is 0, means the binary is loaded at its default base address.
            veritesting (bool): Enable veritesting. Default is False.
            kwargs (dict): Additional keyword arguments to pass to the angr project.
        Short name: l
        """
        try:
            if veritesting:
                kwargs['veritesting'] = True
            if not os.path.exists(binary_path):
                raise DebuggerCommandError(f"File '{binary_path}' not found.")
            if base_addr:
                kwargs['base_addr'] = base_addr
            self.debugger.init(binary_path, **kwargs)
        except Exception as e:
            raise DebuggerCommandError(f"Failed to load binary: {e}")
        f = os.path.basename(binary_path)
        self.send_info(f"Binary '{f}' loaded.")

    def set_entry_state(self, addr:int|None=None, *args, **kwargs):
        """
        Set the call state of a function.

        Args:
            addr (int): The address the state should start at instead of the entry point.
            args (tuple): a list of values to use as the program's argv. May be mixed strings and bitvectors.
            kwargs (dict): a dictionary of additional keyword arguments to pass to the entry_state function.

        Raises:
            DebuggerCommandError: If the function address is not found.
        
        Short name: ses
        """
        self.debugger.set_entry_state(addr, *args, **kwargs)
        self.send_info(f"Execution will start {'at address '+hex(addr) if addr else 'at specified entry point'}.")
    def set_full_state(self, *args, **kwargs):
        """
        Set a full state.

        Args:
            args (tuple): a list of values to use as the program's argv. May be mixed strings and bitvectors.
            kwargs (dict): a dictionary of additional keyword arguments to pass to the entry_state function.

        Short name: sfs
        """
        self.debugger.set_full_state(*args, **kwargs)
        self.send_info(f"Full state set.")
        
    def set_blank_state(self, *args, **kwargs):
        """
        Set a blank state.

        Args:
            args (tuple): a list of values to use as the program's argv. May be mixed strings and bitvectors.
            kwargs (dict): a dictionary of additional keyword arguments to pass to the entry_state function.

        Short name: sbs
        """
        self.debugger.set_blank_state(*args, **kwargs)
        self.send_info(f"Blank state set.")

    def get_current_state(self):
        """
        Get the current state.

        Short name: gcs
        """
        return self.debugger.current_state

    def keep_unconstrained(self, keep:bool=True):
        """
        Keep unconstrained states.

        Args:
            keep (bool): True to keep unconstrained states, False otherwise.

        Short name: ku
        """
        self.debugger.keep_unconstrained = keep
        self.send_info(f"Unconstrained states {'kept' if keep else 'discarded'}.")

    def pause(self):
        """
        Pause the debugger.

        Short name: p
        """
        self.send_info("Paused successfully.")

    def back(self):
        """
        Go back to the previous state.

        Short name: b
        """
        self.debugger.back()

        self.send_info("Stepped back to: " + hex(self.debugger.current_state.addr) +".") # type: ignore

    def reset_state(self):
        """
        Restart the debugger to the entry state.

        Short name: r
        """
        self.debugger.reset_state()
        entry_point = self.debugger.entry_point
        if entry_point is None:
            self.send_info("State reset.")
        else:
            s = f"address {hex(entry_point)}" if isinstance(entry_point, int) else f"to function {entry_point[0]} with arguments {[str(a) for a in entry_point[3]]}"
            self.send_info(f"State reset at entry point {s}.")


    def run_script(self, script_path:str):
        """
        Run a script.

        Args:
            script_path (str): The path to the script to run.

        Short name: dangr
        """
        # read the script and call handler for each non-empty, non-comment line
        try:
            log.debug(lambda: f"running script {script_path} in folder" + os.getcwd())
            for line in ScriptProcessor(script_path).process_file():
                if not line:
                    continue
                if not line.strip().startswith("#"):
                    log.debug(lambda: f"running following line: {line}")
                    if line == "less":
                        continue
                    if not self.debugger.handle(line):
                        break
        except Exception as e:
            raise DebuggerCommandError(f"Failed to run script: {e}", e)


    def select_state(self, index:int|angr.SimState, stash:str="active"):
        """
        Select a path to execute.

        Args:
            index (int|angr.SimState): The index of the path to select as shown in list_active_paths.
            stash (str): The stash to select the path from. Default is active.
        
        Short name: sp
        """
        if isinstance(index, angr.SimState):
            self.debugger.current_state = index
        else:
            self.debugger.set_current_state(index, stash)
        try:
            addr = f": {hex(self.debugger.current_state.addr)}" # type: ignore
        except angr.errors.SimValueError as e:
            addr = ""

        self.send_info(f"Path {index} selected{addr}")

    def move_state_to_stash(self, index:int, from_stash:str, to_stash:str):
        """
        Move a state from one stash to another.

        Args:
            index (int): The index of the state to move.
            from_stash (str): The stash to move the state from.
            to_stash (str): The stash to move the state to.

        Short name: msts
        """
        self.debugger.move_state_to_stash(index, from_stash, to_stash)
        self.send_info(f"State moved from {from_stash} to {to_stash}.")
    
    def move_to_stash(self, stash:str):
        """
        Move a state to a stash.

        Args:
            stash (str): The stash to move the state to.

        Short name: mts
        """
        self.debugger.to_stash(stash)
        self.send_info(f"Current active state moved to {stash}.")


# Generated from /workspaces/dAngr/src/dAngr/cli/grammar/ranges.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,61,59,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,1,0,1,0,1,1,1,1,1,1,3,1,24,8,1,1,2,1,2,1,2,1,2,
        1,2,1,3,1,3,1,3,1,3,1,3,1,4,1,4,1,4,1,4,1,4,1,5,1,5,1,6,1,6,1,7,
        1,7,1,7,1,7,1,7,1,7,1,7,1,7,1,7,1,7,3,7,55,8,7,1,8,1,8,1,8,0,0,9,
        0,2,4,6,8,10,12,14,16,0,1,3,0,20,34,36,58,60,61,57,0,18,1,0,0,0,
        2,23,1,0,0,0,4,25,1,0,0,0,6,30,1,0,0,0,8,35,1,0,0,0,10,40,1,0,0,
        0,12,42,1,0,0,0,14,54,1,0,0,0,16,56,1,0,0,0,18,19,3,2,1,0,19,1,1,
        0,0,0,20,24,3,4,2,0,21,24,3,6,3,0,22,24,3,8,4,0,23,20,1,0,0,0,23,
        21,1,0,0,0,23,22,1,0,0,0,24,3,1,0,0,0,25,26,5,21,0,0,26,27,5,18,
        0,0,27,28,3,0,0,0,28,29,5,19,0,0,29,5,1,0,0,0,30,31,5,22,0,0,31,
        32,5,18,0,0,32,33,3,10,5,0,33,34,5,19,0,0,34,7,1,0,0,0,35,36,5,20,
        0,0,36,37,5,18,0,0,37,38,3,12,6,0,38,39,5,19,0,0,39,9,1,0,0,0,40,
        41,3,14,7,0,41,11,1,0,0,0,42,43,3,14,7,0,43,13,1,0,0,0,44,55,5,4,
        0,0,45,55,5,2,0,0,46,55,3,16,8,0,47,55,5,11,0,0,48,55,5,12,0,0,49,
        55,5,60,0,0,50,51,5,18,0,0,51,52,3,14,7,0,52,53,5,19,0,0,53,55,1,
        0,0,0,54,44,1,0,0,0,54,45,1,0,0,0,54,46,1,0,0,0,54,47,1,0,0,0,54,
        48,1,0,0,0,54,49,1,0,0,0,54,50,1,0,0,0,55,15,1,0,0,0,56,57,7,0,0,
        0,57,17,1,0,0,0,2,23,54
    ]

class rangesParser ( Parser ):

    grammarFileName = "ranges.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'&sym'", "'&reg'", "'&vars'", 
                     "'&mem'", "'&state'", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "'->'", "'('", 
                     "')'", "'!'", "'&'", "'$'", "':'", "';'", "','", "'\"'", 
                     "'''", "'@'", "'.'", "'|'", "'['", "']'", "'{'", "'}'", 
                     "'^'", "'#'", "'%'", "'*'", "'+'", "'/'", "'//'", "'<<'", 
                     "'>>'", "'**'", "'='", "'=='", "'!='", "'<'", "'>'", 
                     "'<='", "'>='", "'&&'", "'||'", "'?'", "'~'", "'`'", 
                     "'_'", "'-'" ]

    symbolicNames = [ "<INVALID>", "HEX_NUMBERS", "NUMBERS", "NUMBER", "LETTERS", 
                      "LETTER", "SYM_DB", "REG_DB", "VARS_DB", "MEM_DB", 
                      "STATE", "STRING", "BINARY_STRING", "ESCAPED_QUOTE", 
                      "ESCAPED_SINGLE_QUOTE", "SESC_SEQ", "ESC_SEQ", "ARROW", 
                      "LPAREN", "RPAREN", "BANG", "AMP", "DOLLAR", "COLON", 
                      "SCOLON", "COMMA", "QUOTE", "SQUOTE", "AT", "DOT", 
                      "BAR", "BRA", "KET", "BRACE", "KETCE", "XOR", "HASH", 
                      "PERC", "MUL", "ADD", "DIV", "FLOORDIV", "LSHIFT", 
                      "RSHIFT", "POW", "ASSIGN", "EQ", "NEQ", "LT", "GT", 
                      "LE", "GE", "AND", "OR", "QMARK", "TILDE", "TICK", 
                      "UNDERSCORE", "DASH", "NEWLINE", "WS", "HAT" ]

    RULE_expression = 0
    RULE_range = 1
    RULE_dangr_range = 2
    RULE_bash_range = 3
    RULE_python_range = 4
    RULE_bash_content = 5
    RULE_py_content = 6
    RULE_anything = 7
    RULE_symbol = 8

    ruleNames =  [ "expression", "range", "dangr_range", "bash_range", "python_range", 
                   "bash_content", "py_content", "anything", "symbol" ]

    EOF = Token.EOF
    HEX_NUMBERS=1
    NUMBERS=2
    NUMBER=3
    LETTERS=4
    LETTER=5
    SYM_DB=6
    REG_DB=7
    VARS_DB=8
    MEM_DB=9
    STATE=10
    STRING=11
    BINARY_STRING=12
    ESCAPED_QUOTE=13
    ESCAPED_SINGLE_QUOTE=14
    SESC_SEQ=15
    ESC_SEQ=16
    ARROW=17
    LPAREN=18
    RPAREN=19
    BANG=20
    AMP=21
    DOLLAR=22
    COLON=23
    SCOLON=24
    COMMA=25
    QUOTE=26
    SQUOTE=27
    AT=28
    DOT=29
    BAR=30
    BRA=31
    KET=32
    BRACE=33
    KETCE=34
    XOR=35
    HASH=36
    PERC=37
    MUL=38
    ADD=39
    DIV=40
    FLOORDIV=41
    LSHIFT=42
    RSHIFT=43
    POW=44
    ASSIGN=45
    EQ=46
    NEQ=47
    LT=48
    GT=49
    LE=50
    GE=51
    AND=52
    OR=53
    QMARK=54
    TILDE=55
    TICK=56
    UNDERSCORE=57
    DASH=58
    NEWLINE=59
    WS=60
    HAT=61

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ExpressionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def range_(self):
            return self.getTypedRuleContext(rangesParser.RangeContext,0)


        def getRuleIndex(self):
            return rangesParser.RULE_expression

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExpression" ):
                listener.enterExpression(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExpression" ):
                listener.exitExpression(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpression" ):
                return visitor.visitExpression(self)
            else:
                return visitor.visitChildren(self)




    def expression(self):

        localctx = rangesParser.ExpressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_expression)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 18
            self.range_()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RangeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dangr_range(self):
            return self.getTypedRuleContext(rangesParser.Dangr_rangeContext,0)


        def bash_range(self):
            return self.getTypedRuleContext(rangesParser.Bash_rangeContext,0)


        def python_range(self):
            return self.getTypedRuleContext(rangesParser.Python_rangeContext,0)


        def getRuleIndex(self):
            return rangesParser.RULE_range

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRange" ):
                listener.enterRange(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRange" ):
                listener.exitRange(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRange" ):
                return visitor.visitRange(self)
            else:
                return visitor.visitChildren(self)




    def range_(self):

        localctx = rangesParser.RangeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_range)
        try:
            self.state = 23
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [21]:
                self.enterOuterAlt(localctx, 1)
                self.state = 20
                self.dangr_range()
                pass
            elif token in [22]:
                self.enterOuterAlt(localctx, 2)
                self.state = 21
                self.bash_range()
                pass
            elif token in [20]:
                self.enterOuterAlt(localctx, 3)
                self.state = 22
                self.python_range()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dangr_rangeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def AMP(self):
            return self.getToken(rangesParser.AMP, 0)

        def LPAREN(self):
            return self.getToken(rangesParser.LPAREN, 0)

        def expression(self):
            return self.getTypedRuleContext(rangesParser.ExpressionContext,0)


        def RPAREN(self):
            return self.getToken(rangesParser.RPAREN, 0)

        def getRuleIndex(self):
            return rangesParser.RULE_dangr_range

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDangr_range" ):
                listener.enterDangr_range(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDangr_range" ):
                listener.exitDangr_range(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDangr_range" ):
                return visitor.visitDangr_range(self)
            else:
                return visitor.visitChildren(self)




    def dangr_range(self):

        localctx = rangesParser.Dangr_rangeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_dangr_range)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 25
            self.match(rangesParser.AMP)
            self.state = 26
            self.match(rangesParser.LPAREN)
            self.state = 27
            self.expression()
            self.state = 28
            self.match(rangesParser.RPAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Bash_rangeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOLLAR(self):
            return self.getToken(rangesParser.DOLLAR, 0)

        def LPAREN(self):
            return self.getToken(rangesParser.LPAREN, 0)

        def bash_content(self):
            return self.getTypedRuleContext(rangesParser.Bash_contentContext,0)


        def RPAREN(self):
            return self.getToken(rangesParser.RPAREN, 0)

        def getRuleIndex(self):
            return rangesParser.RULE_bash_range

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBash_range" ):
                listener.enterBash_range(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBash_range" ):
                listener.exitBash_range(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitBash_range" ):
                return visitor.visitBash_range(self)
            else:
                return visitor.visitChildren(self)




    def bash_range(self):

        localctx = rangesParser.Bash_rangeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_bash_range)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self.match(rangesParser.DOLLAR)
            self.state = 31
            self.match(rangesParser.LPAREN)
            self.state = 32
            self.bash_content()
            self.state = 33
            self.match(rangesParser.RPAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Python_rangeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BANG(self):
            return self.getToken(rangesParser.BANG, 0)

        def LPAREN(self):
            return self.getToken(rangesParser.LPAREN, 0)

        def py_content(self):
            return self.getTypedRuleContext(rangesParser.Py_contentContext,0)


        def RPAREN(self):
            return self.getToken(rangesParser.RPAREN, 0)

        def getRuleIndex(self):
            return rangesParser.RULE_python_range

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPython_range" ):
                listener.enterPython_range(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPython_range" ):
                listener.exitPython_range(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPython_range" ):
                return visitor.visitPython_range(self)
            else:
                return visitor.visitChildren(self)




    def python_range(self):

        localctx = rangesParser.Python_rangeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_python_range)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 35
            self.match(rangesParser.BANG)
            self.state = 36
            self.match(rangesParser.LPAREN)
            self.state = 37
            self.py_content()
            self.state = 38
            self.match(rangesParser.RPAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Bash_contentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def anything(self):
            return self.getTypedRuleContext(rangesParser.AnythingContext,0)


        def getRuleIndex(self):
            return rangesParser.RULE_bash_content

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBash_content" ):
                listener.enterBash_content(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBash_content" ):
                listener.exitBash_content(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitBash_content" ):
                return visitor.visitBash_content(self)
            else:
                return visitor.visitChildren(self)




    def bash_content(self):

        localctx = rangesParser.Bash_contentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_bash_content)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.anything()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Py_contentContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def anything(self):
            return self.getTypedRuleContext(rangesParser.AnythingContext,0)


        def getRuleIndex(self):
            return rangesParser.RULE_py_content

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPy_content" ):
                listener.enterPy_content(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPy_content" ):
                listener.exitPy_content(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPy_content" ):
                return visitor.visitPy_content(self)
            else:
                return visitor.visitChildren(self)




    def py_content(self):

        localctx = rangesParser.Py_contentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_py_content)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 42
            self.anything()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AnythingContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LETTERS(self):
            return self.getToken(rangesParser.LETTERS, 0)

        def NUMBERS(self):
            return self.getToken(rangesParser.NUMBERS, 0)

        def symbol(self):
            return self.getTypedRuleContext(rangesParser.SymbolContext,0)


        def STRING(self):
            return self.getToken(rangesParser.STRING, 0)

        def BINARY_STRING(self):
            return self.getToken(rangesParser.BINARY_STRING, 0)

        def WS(self):
            return self.getToken(rangesParser.WS, 0)

        def LPAREN(self):
            return self.getToken(rangesParser.LPAREN, 0)

        def anything(self):
            return self.getTypedRuleContext(rangesParser.AnythingContext,0)


        def RPAREN(self):
            return self.getToken(rangesParser.RPAREN, 0)

        def getRuleIndex(self):
            return rangesParser.RULE_anything

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAnything" ):
                listener.enterAnything(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAnything" ):
                listener.exitAnything(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAnything" ):
                return visitor.visitAnything(self)
            else:
                return visitor.visitChildren(self)




    def anything(self):

        localctx = rangesParser.AnythingContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_anything)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 54
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.state = 44
                self.match(rangesParser.LETTERS)
                pass

            elif la_ == 2:
                self.state = 45
                self.match(rangesParser.NUMBERS)
                pass

            elif la_ == 3:
                self.state = 46
                self.symbol()
                pass

            elif la_ == 4:
                self.state = 47
                self.match(rangesParser.STRING)
                pass

            elif la_ == 5:
                self.state = 48
                self.match(rangesParser.BINARY_STRING)
                pass

            elif la_ == 6:
                self.state = 49
                self.match(rangesParser.WS)
                pass

            elif la_ == 7:
                self.state = 50
                self.match(rangesParser.LPAREN)
                self.state = 51
                self.anything()
                self.state = 52
                self.match(rangesParser.RPAREN)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SymbolContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WS(self):
            return self.getToken(rangesParser.WS, 0)

        def BANG(self):
            return self.getToken(rangesParser.BANG, 0)

        def AMP(self):
            return self.getToken(rangesParser.AMP, 0)

        def DOLLAR(self):
            return self.getToken(rangesParser.DOLLAR, 0)

        def COLON(self):
            return self.getToken(rangesParser.COLON, 0)

        def SCOLON(self):
            return self.getToken(rangesParser.SCOLON, 0)

        def COMMA(self):
            return self.getToken(rangesParser.COMMA, 0)

        def QUOTE(self):
            return self.getToken(rangesParser.QUOTE, 0)

        def SQUOTE(self):
            return self.getToken(rangesParser.SQUOTE, 0)

        def AT(self):
            return self.getToken(rangesParser.AT, 0)

        def DOT(self):
            return self.getToken(rangesParser.DOT, 0)

        def BAR(self):
            return self.getToken(rangesParser.BAR, 0)

        def BRA(self):
            return self.getToken(rangesParser.BRA, 0)

        def KET(self):
            return self.getToken(rangesParser.KET, 0)

        def BRACE(self):
            return self.getToken(rangesParser.BRACE, 0)

        def KETCE(self):
            return self.getToken(rangesParser.KETCE, 0)

        def HAT(self):
            return self.getToken(rangesParser.HAT, 0)

        def HASH(self):
            return self.getToken(rangesParser.HASH, 0)

        def PERC(self):
            return self.getToken(rangesParser.PERC, 0)

        def MUL(self):
            return self.getToken(rangesParser.MUL, 0)

        def ADD(self):
            return self.getToken(rangesParser.ADD, 0)

        def DIV(self):
            return self.getToken(rangesParser.DIV, 0)

        def POW(self):
            return self.getToken(rangesParser.POW, 0)

        def ASSIGN(self):
            return self.getToken(rangesParser.ASSIGN, 0)

        def EQ(self):
            return self.getToken(rangesParser.EQ, 0)

        def NEQ(self):
            return self.getToken(rangesParser.NEQ, 0)

        def LT(self):
            return self.getToken(rangesParser.LT, 0)

        def GT(self):
            return self.getToken(rangesParser.GT, 0)

        def LE(self):
            return self.getToken(rangesParser.LE, 0)

        def GE(self):
            return self.getToken(rangesParser.GE, 0)

        def AND(self):
            return self.getToken(rangesParser.AND, 0)

        def OR(self):
            return self.getToken(rangesParser.OR, 0)

        def QMARK(self):
            return self.getToken(rangesParser.QMARK, 0)

        def TILDE(self):
            return self.getToken(rangesParser.TILDE, 0)

        def TICK(self):
            return self.getToken(rangesParser.TICK, 0)

        def UNDERSCORE(self):
            return self.getToken(rangesParser.UNDERSCORE, 0)

        def DASH(self):
            return self.getToken(rangesParser.DASH, 0)

        def FLOORDIV(self):
            return self.getToken(rangesParser.FLOORDIV, 0)

        def LSHIFT(self):
            return self.getToken(rangesParser.LSHIFT, 0)

        def RSHIFT(self):
            return self.getToken(rangesParser.RSHIFT, 0)

        def getRuleIndex(self):
            return rangesParser.RULE_symbol

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSymbol" ):
                listener.enterSymbol(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSymbol" ):
                listener.exitSymbol(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSymbol" ):
                return visitor.visitSymbol(self)
            else:
                return visitor.visitChildren(self)




    def symbol(self):

        localctx = rangesParser.SymbolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_symbol)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 56
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 4035225231763177472) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection


import pytest


import os
from unittest.mock import Mock


class TestDebugInfoCommands:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)


    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        assert dbg.handle("load example")
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("run")
        return dbg

    # 
    # def test_get_cfg(self, dbg, conn):
    #     assert dbg.handle("get_cfg")
    #     assert r == True
    #     assert "{\'graph\': \'digraph " in str(conn.send_info.call_args[0][0])
    
    
    def test_get_basicblocks(self, dbg, conn):
        assert dbg.handle("get_basicblocks")
        assert "Address: 0x40059b" in str(conn.send_result.call_args[0][0])

    
    def test_get_current_block(self, dbg, conn):
        assert dbg.handle("get_current_block")
        assert "Address: 0x400566:" in str(conn.send_result.call_args[0][0])
    
    
    def test_list_active_paths(self, dbg, conn):
        assert dbg.handle("list_states 'active'")
        assert "[<SimState @ 0x400566>]" == str(conn.send_result.call_args[0][0])    
    
    def test_list_active_paths2(self, dbg, conn):
        assert dbg.handle("list_states")
        assert "[<SimState @ 0x400566>]" == str(conn.send_result.call_args[0][0])    
    
    def test_list_deadended_paths(self, dbg, conn):
        assert dbg.handle("list_states 'deadended'")
        assert "[]" == str(conn.send_result.call_args[0][0])    

    
    def test_list_binary_symbols(self, dbg, conn):
        assert dbg.handle("list_binary_symbols")
        assert "Binary Symbols:" in str(conn.send_result.call_args[0][0])

    
    def test_list_constraints(self, dbg, conn):
        assert dbg.handle("list_constraints")
        assert "Constraints:" in str(conn.send_result.call_args[0][0])
    
    
    def test_list_path_history(self, dbg, conn):
        assert dbg.handle("list_path_history")
        assert "Path History:" in str(conn.send_result.call_args[0][0])
    
    
    def test_list_breakpoints(self, dbg, conn):
        assert dbg.handle("list_breakpoints")
        assert "Breakpoint(s): [0] Address Filter: 0x400566" == str(conn.send_result.call_args[0][0])
    
    



from dAngr.exceptions import DebuggerCommandError
from .base import BaseCommand
from .filters import FilterCommands
from dAngr.cli.filters import AddressFilter, Filter, FunctionFilter

class BreakpointCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)
        
    def add_breakpoint(self, address:int):  # type: ignore
        """
        Set a breakpoint at a given address.
    
        Args:
            address (int): Address to set the breakpoint at.
        
        Short name: ba
        """
        f = FilterCommands(self.debugger).by_address(address)
        self.debugger.breakpoints.append(f)
        self.send_info(f"Address {hex(address)} added to breakpoints.")
    
    def add_breakpoint_at_line(self, source_file: str, line: int):  # type: ignore
        """
        Set a breakpoint at a given line of a source file. Requires debug symbols to be loaded.
    
        Args:
            source_file (str): Source file to set the breakpoint at.
            line (int): Line number to set the breakpoint at.
        
        Short name: bal
        """
        f = FilterCommands(self.debugger).by_line(source_file, line)
        self.debugger.breakpoints.append(f)
        self.send_info(f"Address {hex(f.address)} added to breakpoints.")
    
    def add_breakpoint_at_function(self, function: str):  # type: ignore
        """
        Set a breakpoint at a given function. Requires debug symbols to be loaded.
    
        Args:
            function (str): Function to set the breakpoint at.
        
        Short name: baf
        """
        f = FilterCommands(self.debugger).by_function(function)
        self.debugger.breakpoints.append(f)
        self.send_info(f"Function {function} added to breakpoints.")
    
    def remove_breakpoint(self, address:int):  # type: ignore
        """
        Remove a breakpoint at a given address.
    
        Args:
            address (int): Address to remove the breakpoint from.
        
        Short name: bra
        """

        f = self.debugger.breakpoints.find(AddressFilter, lambda f: f.address == address) # type: ignore
        if f is None:
            self.send_error(f"Breakpoint at address {hex(address)} not found.")
            return
        self.debugger.breakpoints.remove(f)
        self.send_info(f"Address {hex(address)} removed from breakpoints.")

    
    def remove_breakpoint_at_line(self, source_file: str, line: int):  # type: ignore
        """
        Remove a breakpoint at a given line of a source file. Requires debug symbols to be loaded.
    
        Args:
            source_file (str): Source file to remove the breakpoint from.
            line (int): Line number to remove the breakpoint from.
        
        Short name: brl
        """
        address = self.debugger.find_address(source_file, line)
        if address is None:
            self.send_error(f"Breakpoint at {source_file}:{line} not found.")
            return
        f = self.debugger.breakpoints.find(FunctionFilter, lambda f: f.f_addr == address)
        if f is None:
            self.send_error(f"Breakpoint at {source_file}:{line} not found.")
            return
        self.debugger.breakpoints.remove(f)
        self.send_info(f"Address {hex(address)} removed from breakpoints.")
    
    def remove_breakpoint_at_function(self, function: str):  # type: ignore
        """
        Remove a breakpoint at a given function. Requires debug symbols to be loaded.
    
        Args:
            function (str): Function to remove the breakpoint from.
        
        Short name: brfun
        """
        address = self.debugger.get_function_address(function)
        f = self.debugger.breakpoints.find(FunctionFilter, lambda f: f.f_addr == address)
        if f is None:
            self.send_error(f"Breakpoint at function {function} not found.")
            return
        self.debugger.breakpoints.remove(f)
        self.send_info(f"Function {function} removed from breakpoints.")

    def enable_breakpoint(self, index:int=0, enable:bool=True):
        """
        Enable filter at given index.

        Args:
            index (int): Index of the filter found using list_breakpoints.
            enable (bool): True to enable, False to disable.
        
        Short name: be
        """
        list = self.debugger.breakpoints
        if index >= len(list):
            raise DebuggerCommandError(f"Index {index} out of range.")
        list[index].enabled = enable
        self.send_info(f"Breakpoint filter {'enabled' if enable else 'disabled'}.")
    def disable_breakpoint(self, index:int=0):
        """
        Disable filter at given index.

        Args:
            index (int): Index of the filter found using list_breakpoints.
        
        Short name: bd
        """
        list = self.debugger.breakpoints
        if index >= len(list):
            raise DebuggerCommandError(f"Index {index} out of range.")
        list[index].enabled = False
        self.send_info(f"Breakpoint filter disabled.")

    def list_breakpoints(self):
        """
        List all breakpoints.
        
        Short name: bl
        """
        list = self.debugger.breakpoints
        if len(list) == 0:
            self.send_info(f'No breakpoints found.')
            return []
        return f'Breakpoint(s): {"\n\t".join([f"[{i}] {b}" for i,b in enumerate(list)])}'
    
    def breakpoint(self, *filters:Filter):
            """
            Add a filter to the list of breakpoints.

            Args:
                filters (tuple): The filters to add. Can be an address, source file and line number, function name or stream text.
            
            Short name: bf
            """

            FilterCommands(self.debugger).filter(False, *filters)

    def remove_breakpoint_filter(self, index:int):
        """
        Remove a filter from the list of breakpoints.

        Args:
            index (int): Index of the filter found using list_filters.
        
        Short name: brf
        """
        FilterCommands(self.debugger).remove_filter(index, False)

    def clear_breakpoints(self):
        """
        Clear all breakpoints.

        Short name: bc
        """
        self.debugger.breakpoints.clear()
        self.send_info( "All breakpoints cleared.")


from typing import override
from antlr4.error.ErrorListener import ConsoleErrorListener

class ErrorListener(ConsoleErrorListener):
    def __init__(self, name):
        super(ErrorListener, self).__init__()
        self.name = name
        self._errors = []
    
    @override
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        error_message = f"{self.name} at line {line}, column {column}: {msg}"
        self._errors.append(error_message)
        # print(error_message)
    
    @property
    def errors(self):
        return self._errors
    
    def reportErrors(self):
        if self._errors:
            print(f"Found {len(self._errors)} error(s):")
            for error in self._errors:
                print(error)
        else:
            print("No syntax errors found.")




from typing import Any, List
from antlr4 import CommonTokenStream, InputStream, Lexer, Token
from dAngr.cli.grammar.antlr.dAngrLexer import dAngrLexer
from dAngr.cli.grammar.antlr.dAngrParser import dAngrParser
from dAngr.cli.grammar.error_listener import ErrorListener
from dAngr.cli.grammar.visitor import dAngrVisitor_
from dAngr.cli.grammar.script import Script
from dAngr.exceptions import ParseError

from dAngr.utils.loggers import AsyncLogger

log = AsyncLogger(__file__)


def logTokens(stream:CommonTokenStream, Parser:type[Any], lexer:Any):
    stream.fill()
    result:List[str] = []
    line = -1
    # print token types and value per line, hence if line number changes, print on new line

    switcher = {
                -1: "EOF",
    }
    if getattr(Parser, "INDENT", None):
        switcher[Parser.INDENT] = "INDENT"
        switcher[Parser.DEDENT] = "DEDENT"

    for token in stream.tokens:
        if token.line != line:
            line = token.line
            result.append(f"{token.line}: ")
        if token.type in switcher:
            token_type = switcher[token.type]
        else:
            token_type = lexer.ruleNames[token.type-1]
        result[-1] +=f"{token_type}({token.text}) "

    log.debug(lambda:"\n".join(result))
    return result

def lex_input(input, Lexer:type[Any]=dAngrLexer, Parser:type[Any]=dAngrParser):
    input_stream = InputStream(input)
    lexer = Lexer(input_stream)
    stream = CommonTokenStream(lexer)
    error_listener = ErrorListener("Parse error")
    lexer.addErrorListener(error_listener)
    if error_listener._errors:
        raise ParseError("\n".join([e for e in error_listener.errors]))
    return logTokens(stream, Parser, lexer)

def validate_input(input:str):
    input_stream = InputStream(input)
    lexer = dAngrLexer(input_stream)
    lexer.removeErrorListeners()
    syntax_error_listener = ErrorListener("Syntax error")
    lexer.addErrorListener(syntax_error_listener)
    stream = CommonTokenStream(lexer)
    parser = dAngrParser(stream)
    error_listener = ErrorListener("Parse error")
    parser.removeErrorListeners()  # Remove default console error listener
    parser.addErrorListener(error_listener)
    tree = parser.script()
    if syntax_error_listener._errors:
        return str(syntax_error_listener.errors[0])
    if error_listener._errors:
       return str(error_listener.errors[0])
    if not tree:
        return "No tree generated"
    

def parse_input(input:str, debugger, Lexer:type[Any]=dAngrLexer, Parser:type[Any] = dAngrParser, Visitor:type[Any]|None = dAngrVisitor_)-> Script:
    input_stream = InputStream(input)
    lexer = Lexer(input_stream)
    lexer.removeErrorListeners()
    syntax_error_listener = ErrorListener("Syntax error")
    lexer.addErrorListener(syntax_error_listener)
    stream = CommonTokenStream(lexer)
    logTokens(stream, Parser, lexer)
    parser = Parser(stream)
    error_listener = ErrorListener("Parse error")
    parser.removeErrorListeners()  # Remove default console error listener
    parser.addErrorListener(error_listener)
    tree = parser.script()
    if syntax_error_listener._errors:
        raise ParseError("\n".join([e for e in syntax_error_listener.errors]))
    if error_listener._errors:
        raise ParseError("\n".join([e for e in error_listener.errors]))
    if not tree:
        raise ParseError("No tree generated")
    log.debug(lambda:tree.toStringTree(recog=parser))
    
    if not Visitor:
        raise ParseError("No visitor provided")
    visitor = Visitor(debugger)
    return visitor.visit(tree)




import claripy
from dAngr.exceptions import DebuggerCommandError
from dAngr.utils.utils import StreamType, SymBitVector, Variable, convert_argument
from .base import BaseCommand

class FileCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)

    def dump_stdstream(self, stream_type: StreamType = StreamType.stdout):
        """
        Dump the standard stream.

        Args:
            stream_type (StreamType): The type of the stream to dump. Default is stdout.
        
        Short name: ds
        
        """
        return self.debugger.get_stdstream(stream_type)


    def create_symbolic_file(self, name:str, content:str|SymBitVector|Variable|None=None, size:int|None=None):
        """
        Create a symbolic file with name and size.

        Args:
            name (str): Name of the file
            content (str|SymBitVector|Variable|None): Content of the file. Default is None. If content starts with $sym.SYM, the content is replaced with the symbol named SYM.
            size (int|None): Size of the file. Default is None.
        
        Short name: csf
        
        """
        c = None
        if not content is None:
            c = self.to_value(content)
        self.debugger.create_symbolic_file(name, c, size) # type: ignore
        self.send_info(f"Symbolic file {name} created.")


# Generated from /workspaces/dAngr/src/dAngr/cli/grammar/dAngr.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .dAngrParser import dAngrParser
else:
    from dAngrParser import dAngrParser

# This class defines a complete listener for a parse tree produced by dAngrParser.
class dAngrListener(ParseTreeListener):

    # Enter a parse tree produced by dAngrParser#script.
    def enterScript(self, ctx:dAngrParser.ScriptContext):
        pass

    # Exit a parse tree produced by dAngrParser#script.
    def exitScript(self, ctx:dAngrParser.ScriptContext):
        pass


    # Enter a parse tree produced by dAngrParser#statement.
    def enterStatement(self, ctx:dAngrParser.StatementContext):
        pass

    # Exit a parse tree produced by dAngrParser#statement.
    def exitStatement(self, ctx:dAngrParser.StatementContext):
        pass


    # Enter a parse tree produced by dAngrParser#expression.
    def enterExpression(self, ctx:dAngrParser.ExpressionContext):
        pass

    # Exit a parse tree produced by dAngrParser#expression.
    def exitExpression(self, ctx:dAngrParser.ExpressionContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionRange.
    def enterExpressionRange(self, ctx:dAngrParser.ExpressionRangeContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionRange.
    def exitExpressionRange(self, ctx:dAngrParser.ExpressionRangeContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionIn.
    def enterExpressionIn(self, ctx:dAngrParser.ExpressionInContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionIn.
    def exitExpressionIn(self, ctx:dAngrParser.ExpressionInContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionNot.
    def enterExpressionNot(self, ctx:dAngrParser.ExpressionNotContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionNot.
    def exitExpressionNot(self, ctx:dAngrParser.ExpressionNotContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionObject.
    def enterExpressionObject(self, ctx:dAngrParser.ExpressionObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionObject.
    def exitExpressionObject(self, ctx:dAngrParser.ExpressionObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionBool.
    def enterExpressionBool(self, ctx:dAngrParser.ExpressionBoolContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionBool.
    def exitExpressionBool(self, ctx:dAngrParser.ExpressionBoolContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionReference.
    def enterExpressionReference(self, ctx:dAngrParser.ExpressionReferenceContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionReference.
    def exitExpressionReference(self, ctx:dAngrParser.ExpressionReferenceContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionIf.
    def enterExpressionIf(self, ctx:dAngrParser.ExpressionIfContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionIf.
    def exitExpressionIf(self, ctx:dAngrParser.ExpressionIfContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionAlt.
    def enterExpressionAlt(self, ctx:dAngrParser.ExpressionAltContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionAlt.
    def exitExpressionAlt(self, ctx:dAngrParser.ExpressionAltContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionParenthesis.
    def enterExpressionParenthesis(self, ctx:dAngrParser.ExpressionParenthesisContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionParenthesis.
    def exitExpressionParenthesis(self, ctx:dAngrParser.ExpressionParenthesisContext):
        pass


    # Enter a parse tree produced by dAngrParser#ExpressionOperation.
    def enterExpressionOperation(self, ctx:dAngrParser.ExpressionOperationContext):
        pass

    # Exit a parse tree produced by dAngrParser#ExpressionOperation.
    def exitExpressionOperation(self, ctx:dAngrParser.ExpressionOperationContext):
        pass


    # Enter a parse tree produced by dAngrParser#assignment.
    def enterAssignment(self, ctx:dAngrParser.AssignmentContext):
        pass

    # Exit a parse tree produced by dAngrParser#assignment.
    def exitAssignment(self, ctx:dAngrParser.AssignmentContext):
        pass


    # Enter a parse tree produced by dAngrParser#static_var.
    def enterStatic_var(self, ctx:dAngrParser.Static_varContext):
        pass

    # Exit a parse tree produced by dAngrParser#static_var.
    def exitStatic_var(self, ctx:dAngrParser.Static_varContext):
        pass


    # Enter a parse tree produced by dAngrParser#ext_command.
    def enterExt_command(self, ctx:dAngrParser.Ext_commandContext):
        pass

    # Exit a parse tree produced by dAngrParser#ext_command.
    def exitExt_command(self, ctx:dAngrParser.Ext_commandContext):
        pass


    # Enter a parse tree produced by dAngrParser#control_flow.
    def enterControl_flow(self, ctx:dAngrParser.Control_flowContext):
        pass

    # Exit a parse tree produced by dAngrParser#control_flow.
    def exitControl_flow(self, ctx:dAngrParser.Control_flowContext):
        pass


    # Enter a parse tree produced by dAngrParser#else_.
    def enterElse_(self, ctx:dAngrParser.Else_Context):
        pass

    # Exit a parse tree produced by dAngrParser#else_.
    def exitElse_(self, ctx:dAngrParser.Else_Context):
        pass


    # Enter a parse tree produced by dAngrParser#function_def.
    def enterFunction_def(self, ctx:dAngrParser.Function_defContext):
        pass

    # Exit a parse tree produced by dAngrParser#function_def.
    def exitFunction_def(self, ctx:dAngrParser.Function_defContext):
        pass


    # Enter a parse tree produced by dAngrParser#body.
    def enterBody(self, ctx:dAngrParser.BodyContext):
        pass

    # Exit a parse tree produced by dAngrParser#body.
    def exitBody(self, ctx:dAngrParser.BodyContext):
        pass


    # Enter a parse tree produced by dAngrParser#fstatement.
    def enterFstatement(self, ctx:dAngrParser.FstatementContext):
        pass

    # Exit a parse tree produced by dAngrParser#fstatement.
    def exitFstatement(self, ctx:dAngrParser.FstatementContext):
        pass


    # Enter a parse tree produced by dAngrParser#iterable.
    def enterIterable(self, ctx:dAngrParser.IterableContext):
        pass

    # Exit a parse tree produced by dAngrParser#iterable.
    def exitIterable(self, ctx:dAngrParser.IterableContext):
        pass


    # Enter a parse tree produced by dAngrParser#parameters.
    def enterParameters(self, ctx:dAngrParser.ParametersContext):
        pass

    # Exit a parse tree produced by dAngrParser#parameters.
    def exitParameters(self, ctx:dAngrParser.ParametersContext):
        pass


    # Enter a parse tree produced by dAngrParser#condition.
    def enterCondition(self, ctx:dAngrParser.ConditionContext):
        pass

    # Exit a parse tree produced by dAngrParser#condition.
    def exitCondition(self, ctx:dAngrParser.ConditionContext):
        pass


    # Enter a parse tree produced by dAngrParser#operation.
    def enterOperation(self, ctx:dAngrParser.OperationContext):
        pass

    # Exit a parse tree produced by dAngrParser#operation.
    def exitOperation(self, ctx:dAngrParser.OperationContext):
        pass


    # Enter a parse tree produced by dAngrParser#py_basic_content.
    def enterPy_basic_content(self, ctx:dAngrParser.Py_basic_contentContext):
        pass

    # Exit a parse tree produced by dAngrParser#py_basic_content.
    def exitPy_basic_content(self, ctx:dAngrParser.Py_basic_contentContext):
        pass


    # Enter a parse tree produced by dAngrParser#py_content.
    def enterPy_content(self, ctx:dAngrParser.Py_contentContext):
        pass

    # Exit a parse tree produced by dAngrParser#py_content.
    def exitPy_content(self, ctx:dAngrParser.Py_contentContext):
        pass


    # Enter a parse tree produced by dAngrParser#bash_content.
    def enterBash_content(self, ctx:dAngrParser.Bash_contentContext):
        pass

    # Exit a parse tree produced by dAngrParser#bash_content.
    def exitBash_content(self, ctx:dAngrParser.Bash_contentContext):
        pass


    # Enter a parse tree produced by dAngrParser#reference.
    def enterReference(self, ctx:dAngrParser.ReferenceContext):
        pass

    # Exit a parse tree produced by dAngrParser#reference.
    def exitReference(self, ctx:dAngrParser.ReferenceContext):
        pass


    # Enter a parse tree produced by dAngrParser#index.
    def enterIndex(self, ctx:dAngrParser.IndexContext):
        pass

    # Exit a parse tree produced by dAngrParser#index.
    def exitIndex(self, ctx:dAngrParser.IndexContext):
        pass


    # Enter a parse tree produced by dAngrParser#identifier.
    def enterIdentifier(self, ctx:dAngrParser.IdentifierContext):
        pass

    # Exit a parse tree produced by dAngrParser#identifier.
    def exitIdentifier(self, ctx:dAngrParser.IdentifierContext):
        pass


    # Enter a parse tree produced by dAngrParser#numeric.
    def enterNumeric(self, ctx:dAngrParser.NumericContext):
        pass

    # Exit a parse tree produced by dAngrParser#numeric.
    def exitNumeric(self, ctx:dAngrParser.NumericContext):
        pass


    # Enter a parse tree produced by dAngrParser#SlideStartLengthObject.
    def enterSlideStartLengthObject(self, ctx:dAngrParser.SlideStartLengthObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#SlideStartLengthObject.
    def exitSlideStartLengthObject(self, ctx:dAngrParser.SlideStartLengthObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#ReferenceObject.
    def enterReferenceObject(self, ctx:dAngrParser.ReferenceObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#ReferenceObject.
    def exitReferenceObject(self, ctx:dAngrParser.ReferenceObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#BinaryStringObject.
    def enterBinaryStringObject(self, ctx:dAngrParser.BinaryStringObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#BinaryStringObject.
    def exitBinaryStringObject(self, ctx:dAngrParser.BinaryStringObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#ListObject.
    def enterListObject(self, ctx:dAngrParser.ListObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#ListObject.
    def exitListObject(self, ctx:dAngrParser.ListObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#IndexedPropertyObject.
    def enterIndexedPropertyObject(self, ctx:dAngrParser.IndexedPropertyObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#IndexedPropertyObject.
    def exitIndexedPropertyObject(self, ctx:dAngrParser.IndexedPropertyObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#DictionaryObject.
    def enterDictionaryObject(self, ctx:dAngrParser.DictionaryObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#DictionaryObject.
    def exitDictionaryObject(self, ctx:dAngrParser.DictionaryObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#NumericObject.
    def enterNumericObject(self, ctx:dAngrParser.NumericObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#NumericObject.
    def exitNumericObject(self, ctx:dAngrParser.NumericObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#SliceStartEndObject.
    def enterSliceStartEndObject(self, ctx:dAngrParser.SliceStartEndObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#SliceStartEndObject.
    def exitSliceStartEndObject(self, ctx:dAngrParser.SliceStartEndObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#StringObject.
    def enterStringObject(self, ctx:dAngrParser.StringObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#StringObject.
    def exitStringObject(self, ctx:dAngrParser.StringObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#IDObject.
    def enterIDObject(self, ctx:dAngrParser.IDObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#IDObject.
    def exitIDObject(self, ctx:dAngrParser.IDObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#PropertyObject.
    def enterPropertyObject(self, ctx:dAngrParser.PropertyObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#PropertyObject.
    def exitPropertyObject(self, ctx:dAngrParser.PropertyObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#BoolObject.
    def enterBoolObject(self, ctx:dAngrParser.BoolObjectContext):
        pass

    # Exit a parse tree produced by dAngrParser#BoolObject.
    def exitBoolObject(self, ctx:dAngrParser.BoolObjectContext):
        pass


    # Enter a parse tree produced by dAngrParser#anything.
    def enterAnything(self, ctx:dAngrParser.AnythingContext):
        pass

    # Exit a parse tree produced by dAngrParser#anything.
    def exitAnything(self, ctx:dAngrParser.AnythingContext):
        pass


    # Enter a parse tree produced by dAngrParser#anything_no.
    def enterAnything_no(self, ctx:dAngrParser.Anything_noContext):
        pass

    # Exit a parse tree produced by dAngrParser#anything_no.
    def exitAnything_no(self, ctx:dAngrParser.Anything_noContext):
        pass


    # Enter a parse tree produced by dAngrParser#special_words.
    def enterSpecial_words(self, ctx:dAngrParser.Special_wordsContext):
        pass

    # Exit a parse tree produced by dAngrParser#special_words.
    def exitSpecial_words(self, ctx:dAngrParser.Special_wordsContext):
        pass


    # Enter a parse tree produced by dAngrParser#range.
    def enterRange(self, ctx:dAngrParser.RangeContext):
        pass

    # Exit a parse tree produced by dAngrParser#range.
    def exitRange(self, ctx:dAngrParser.RangeContext):
        pass


    # Enter a parse tree produced by dAngrParser#dangr_range.
    def enterDangr_range(self, ctx:dAngrParser.Dangr_rangeContext):
        pass

    # Exit a parse tree produced by dAngrParser#dangr_range.
    def exitDangr_range(self, ctx:dAngrParser.Dangr_rangeContext):
        pass


    # Enter a parse tree produced by dAngrParser#bash_range.
    def enterBash_range(self, ctx:dAngrParser.Bash_rangeContext):
        pass

    # Exit a parse tree produced by dAngrParser#bash_range.
    def exitBash_range(self, ctx:dAngrParser.Bash_rangeContext):
        pass


    # Enter a parse tree produced by dAngrParser#python_range.
    def enterPython_range(self, ctx:dAngrParser.Python_rangeContext):
        pass

    # Exit a parse tree produced by dAngrParser#python_range.
    def exitPython_range(self, ctx:dAngrParser.Python_rangeContext):
        pass


    # Enter a parse tree produced by dAngrParser#symbol.
    def enterSymbol(self, ctx:dAngrParser.SymbolContext):
        pass

    # Exit a parse tree produced by dAngrParser#symbol.
    def exitSymbol(self, ctx:dAngrParser.SymbolContext):
        pass



del dAngrParser



from abc import abstractmethod

from typing import List

from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.cli.grammar.expressions import ReferenceObject

#Statements
class Statement:
    @abstractmethod
    def __call__(self, context):
        raise NotImplementedError
    
    @staticmethod
    def flatten(lst:List['Statement']):
        newList =[]
        for item in lst:
            if isinstance(item, list):
                newList.extend(item)
            else:
                newList.append(item)
        return newList
    @abstractmethod
    def __repr__(self):
        raise NotImplementedError
    @abstractmethod
    def __eq__(self, value: object) -> bool:
        raise NotImplementedError


class Assignment(Statement):
    def __init__(self, variable:ReferenceObject, value):
        self.variable:ReferenceObject = variable
        self.value = value

    def __call__(self, context:ExecutionContext):
        value = self.value(context)
        self.variable.set_value(context, value)
        return None

    def __repr__(self):
        return f"{self.variable} = {self.value}"
    
    def __eq__(self, value: object) -> bool:
        return isinstance(value, Assignment) and self.variable == value.variable and self.value == value.value
    



import argparse
from dAngr.cli.server import Server

def run():
    parser = argparse.ArgumentParser(description="dAngr Symbolic debugger.")
    parser.add_argument("-f", type=str, help="File to debug.", required=False)
    parser.add_argument("-s", type=str, help="Script to execute.",required=False)
    args = parser.parse_args()
    Server(args.f, args.s).start_server()





from abc import abstractmethod
import json

from angr import SimState


class Breakpoint:
    def __init__(self, address, source=None, line_nr=None, enabled=True):
        self.address = address  # The memory address of the breakpoint
        self.source = source    # The source file in which the breakpoint is set
        self.line_nr = line_nr  # The line number of the breakpoint in the source file
        self.enabled = enabled  # Whether the breakpoint is currently active

    def __str__(self):
        status = "" if self.enabled else " (disabled)"
        if self.line_nr is None:
            return f"Breakpoint at {hex(self.address)}{status}"
        else:
            return f"Breakpoint at {hex(self.address)} in {self.source} line  {self.line_nr}{status}"
        
class Register:
    def __init__(self,name,size,offset,value):
        self.name = name
        self.size = size
        self.value = value
        self.offset = offset
    def __repr__(self) -> str:
        v = self.value if not self.value.concrete else hex(self.value.concrete_value)
        return f"Register {self.name} (offset: {hex(self.offset)}, size: {self.size} bytes): {v}"
    
class Memory:
    def __init__(self,address,value,value_type):
        self.address = address
        self.value_type = value_type
        self.value = value
    def __str__(self) -> str:
        return f"Memory at {hex(self.address)}: {self.value} ({self.value_type})"
    
class State:
    def __init__(self,index,address,ended=False):
        self.index:int = index
        self.address:int = address
    def __str__(self) -> str:
        return f"State {self.index} at {hex(self.address)}"
    

class StateExecution:
    def __init__(self,index,output):
        self.index = index
        self.output = output
    def __str__(self) -> str:
        return f"State {self.index} at {self.output}"

class SymbolicVariable:
    def __init__(self,name,value):
        self.name = name
        self.value = value
    def __str__(self) -> str:
        return f"Symvar: {self.name}: {self.value}"




import os
from unittest.mock import Mock
import pytest

from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection
from dAngr.exceptions import CommandError, DebuggerCommandError
from dAngr.exceptions.InvalidArgumentError import InvalidArgumentError

class TestCommands:
    old_dir = os.getcwd()
    
    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)
        
    
    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c
    
    @pytest.fixture
    def dbg(self,conn):
        return CommandLineDebugger(conn)

    
    def test_help(self,dbg, conn):
        assert dbg.handle("help")
        conn.send_result.assert_called_once()
        assert "Available commands:" in conn.send_result.call_args[0][0]
        
    
    def test_help_question_mark(self, dbg, conn):
        assert dbg.handle("?")
        conn.send_result.assert_called_once()
        assert "Available commands:" in conn.send_result.call_args[0][0]

    
    def test_help_command(self, dbg, conn):
        assert dbg.handle("help run")
        conn.send_result.assert_called_once()
        assert "Run until a breakpoint or" in conn.send_result.call_args[0][0]
    
    def test_command_not_found(self, dbg, conn):
        assert dbg.handle("not_a_command", False)
        conn.send_error.assert_called_once()
        assert "Unknown variable:" in str(conn.send_error.call_args[0][0])

    
    def test_exit(self, dbg, conn):
        assert not dbg.handle("exit")
        conn.send_info.assert_not_called()

    
    def test_exit_with_args(self, dbg, conn):
        assert dbg.handle("exit args", False)
        conn.send_info.assert_not_called()
        conn.send_error.assert_called_once_with(InvalidArgumentError('Too many arguments. Expected 0 but got 1'))


    
    def test_load(self, dbg, conn):
        assert dbg.handle("load example")
        conn.send_info.assert_called_once_with("Binary 'example' loaded.")

    
    def test_load_invalid(self, dbg, conn):
        assert dbg.handle("load invalid", False)
        conn.send_error.assert_called_once_with(DebuggerCommandError("Failed to load binary: File 'invalid' not found."))
    
    
    def test_load_args(self, dbg, conn):
        assert dbg.handle("load example args", False)
        conn.send_error.assert_called_once_with(DebuggerCommandError('Unknown variable: args'))
    
    
    def test_variable(self, dbg, conn):
        assert dbg.handle("test = 1")
        assert dbg.handle("&vars.test")
        conn.send_result.assert_called_once()
        assert 1 == conn.send_result.call_args[0][0]


#create a logger that gets lambda functions as input and depending on DEBUG level, logs the output

# The logger should have the following methods:
# - debug
# - info
# - warning
# - error
import logging
from typing import Callable



class AsyncLogger(logging.Logger):
    def __init__(self, name):
        super().__init__(name)

    def debug(self, msg, *args, **kwargs):
        if isinstance(msg, Callable):
            if self.level == logging.DEBUG:
                super().debug(msg(), args, **kwargs)
        else:
            super().debug(msg, *args, **kwargs)
    def info(self, msg, *args, **kwargs):
        if isinstance(msg, Callable):
            if self.level in [logging.DEBUG, logging.INFO]:
                super().info(msg(), *args, **kwargs)
        else:
            super().info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs):
        if isinstance(msg, Callable):
            if self.level == [logging.DEBUG, logging.INFO, logging.WARNING]:
                super().warning(msg(), *args, **kwargs)
        else:
            super().warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        if isinstance(msg, Callable):
            if self.level in [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR]:
                super().error(msg(), *args, **kwargs)
        else:
            super().error(msg, *args, **kwargs)


class LogConfig:
    level_to_str = {
            logging.DEBUG: "DEBUG",
            logging.INFO: "INFO",
            logging.WARNING: "WARNING",
            logging.ERROR: "ERROR",
            logging.CRITICAL: "CRITICAL",
            logging.NOTSET : "NOTSET"
        }
    def __init__(self, default_level: int):
        #set default level
        logging.setLoggerClass(AsyncLogger)
        logging.basicConfig(level=default_level)
        self.settings_per_module = { 
            "angr": logging.ERROR, 
            "pyvex": logging.ERROR, 
            "claripy": logging.ERROR, 
            "cle": logging.ERROR
        }
        for module, level in self.settings_per_module.items():
            self.set_module(module, level)

    def set_module(self, module:str, level: str|int):
        if isinstance(level, str):
            if level not in self.level_to_str.values():
                raise ValueError(f"Invalid log level: {level}")
            level = next(k for k,v in self.level_to_str.items() if v == level)
        self.settings_per_module[module] = level
        logger = logging.getLogger(module)
        logger.setLevel(level)
    
    def list_modules_from_loggers(self):
        
        #retrieve all loggers and their levels as a string
        loggers = logging.Logger.manager.loggerDict
        result = []
        for logger in loggers:
            if isinstance(loggers[logger],logging.Logger):
                l = loggers[logger]
                result.append(f"{logger}: {self.level_to_str[l.level]}") # type: ignore
        #sort the result alfabetically, but put loggers starting with "dAngr" first
        result.sort(key=lambda x: x if x.startswith("dAngr") else f"zzz{x}")
        return result

dAngr_log_config = LogConfig(logging.ERROR)
# dAngr_log_config.set_module("angr", logging.DEBUG)


get_logger = logging.getLogger



import os



from .base import BaseCommand
from dAngr.exceptions.DebuggerCommandError import DebuggerCommandError
from prompt_toolkit.shortcuts import ProgressBar
from prompt_toolkit import ANSI

import angrutils

class InformationCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)

    def get_basicblocks(self):
        """
        Show the assembly for the current basic block.

        Requires Controlflow graph to be geenrated.
        Uses a progress bar to indicate the reconstruction of basic blocks.

        Short name: bbs
        """
        with ProgressBar(title="reconstructing basic blocks") as pb:
            for b in pb(self.debugger.get_bbs()):
                self.send_result(str(b))

    def get_current_block(self):
        """
        Show the current basic block.

        Returns:
            str: The current basic block.

        Raises:
            DebuggerCommandError: If no function is found with the given name.
        
        Short name: bb
        """
        b = self.debugger.get_current_basic_block()
        if b is None:
            raise DebuggerCommandError("No basic block found.")
        return b
    
    def get_cfg(self):
        """
        Show the control flow graph of the current function.

        Returns:
            str: The URL of the control flow graph.
        
        Short name: cfg
        """
        try:
            import pygraphviz as pgv # type: ignore
        except ImportError:
            raise DebuggerCommandError("Please install graphviz, graphviz-dev and pygraphviz to use this command.")

        cfg = self.debugger.cfg
        base_path = self.debugger.launch_file_server()
        file_index = 1 + len([name for name in os.listdir(base_path) if os.path.isfile(os.path.join(base_path, name))])
        filename = "cfg_" + str(file_index)
        svg_path = f"{base_path}/{filename}"
        
        angrutils.plot_cfg(cfg, svg_path, asminst=True, vexinst=False, remove_imports=True, remove_path_terminator=True, format="svg")

        return f"http://localhost:8000/{filename}.svg"

    def verbose_step(self, enable:bool = True):
        """
        Enable or disable verbose stepping.

        Args:
            enable (bool): Enable or disable verbose stepping. Default True.

        Short name: vs
        """
        self.debugger.verbose_step = enable
        return f"Verbose stepping {'enabled' if enable else 'disabled'}."
    
    def get_stashes(self):
        """
        Get the available stashes.

        Returns:
            str: The available stashes.
        
        Short name: gs
        """
        return f"Stashes: {self.debugger.get_stashes()}"

    def list_states(self, stash:str = "active"):
        """
        List all active paths.

        Args:
            stash (str): The stash to use. Default active
        Returns:
            str: Information about the active paths.
        
        Short name: iap
        """
        paths = self.debugger.list_paths(stash) 
        result_list = []
        for i in range(len(paths)):
            result_list.append(f"{i}: " + str(paths[i]))
        return paths



    def list_binary_strings(self, min_length:int = 4):
        """
        List all binary strings.

        Args:
            min_length (int): The minimum length of the binary strings to list. Default 4.

        Returns:
            str: Information about the binary strings.
        
        Short name: ibstr
        """
        strings = self.debugger.get_binary_string_constants(min_length=min_length)
        return f"\taddress\tvalue\n{"\n".join([f"{s[0]}\t{s[1]}" for s in strings])}"

    def list_binary_symbols(self): # type: ignore
        """
        List the debugsymbols when available.

        Returns:
            str: Information about the binary symbols.

        Short name: ibsym
        """
        symbols = self.debugger.get_binary_symbols()
        return f"Binary Symbols: {"\n".join([str(s) for s in symbols])}"

    def list_constraints(self):
        """
        List the current path's constraints and symbolic variables.

        Returns:
            str: Information about the constraints.

        Short name: ic
        """
        ctrs = self.debugger.get_constraints()
        return f"Constraints: {"\n".join([str(c) for c in ctrs])}"

    def list_path_history(self, index:int = 0, stash:str = "active"):
        """
        List the history of a path.

        Args:
            index (int): The index of the path to list the history of. Default active path.
            stash (str): The stash to use. Default active.

        Returns:
            str: Information about the path history.

        Raises:
            DebuggerCommandError: If the path index is invalid.
        
        Short name: iph
        """
        return f"Path History: {"\n".join([str(p) for p in self.debugger.list_path_history(index, stash)])}"
    
    def get_binary_info(self):
        """
        Get information about the binary.

        Returns:
            str: Information about the binary.

        Short name: ibi
        """
        info = self.debugger.get_binary_info()
        return "\n".join([f"{i}: {info[i]}" for i in info])
    
    def get_binary_security_features(self):
        """
        Get security features of the binary.

        Returns:
            str: Information about the security features.

        Short name: ibsf
        """
        info = self.debugger.get_binary_security_features()
        return "\n".join([f"{i}: {info[i]}" for i in info])
    

    def inspect_state(self):
        """
        Inspect the current state.

        Returns:
            str: Information about the current state.

        Short name: is
        """
        # Registers --> refactor list registers to format 'eax': 0x0...
        pstr_state = self.debugger.visulize_state()
        self.send_result(ANSI(pstr_state))
    
# Add current basic block, current function + code
# Name of the symbolic var instead of the to str
# Parcially symbolic...
# Legende


from enum import Enum
import signal
from contextlib import contextmanager

class TimeoutException(Exception): pass

@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class MemLocation(Enum):
    STACK = 0
    HEAP = 1
    CODE = 2
    DATA = 3

class Color(Enum):
    CODE = 31
    HEAP = 33
    SYMBOLIC = 92
    STACK = 35
    YELLOW = 33
    GREEN = 32
    BLUE = 34
    CYAN = 36
    WHITE = 37
    GRAY = 90

def to_color(str:str, color:Color):
    c = color.value
    return f"\x1b[{c}m{str}\x1b[0m"
    
class StateVisualizer():
    REGISTERS = {
    8 : ["al", "ah", "bl", "bh", "cl", "ch", "dl", "dh"],
    16: ["ax", "bx", "cx", "dx"],
    32: ["eax", "ebx", "ecx", "edx", "esi", "edi", "ebp", "esp", "eip"],
    64: ["rax", "rbx", "rcx", "rdx", "rsi", "rdi", "rbp", "rsp", "rip",
        "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15"]}

    def __init__(self, state=None):
        self.state = state
        if state:
            self.bits = state.arch.bits
            self.stack_end = state.registers.load('sp').concrete_value
            bp = state.registers.load('bp')
            if not bp.concrete:
                self.bp = self.stack_end
            else:
                self.bp = bp.concrete_value

            self.stack_begin = state.arch.initial_sp
            
            
            self.heap_start = state.heap.heap_base
            self.heap_end = state.heap.heap_base + state.heap.heap_size
            # state.project.loader.all_objects --> start & end
            self.code_start = state.project.loader.main_object.min_addr
            self.code_end = state.project.loader.main_object.max_addr

    def arch_registers(self, arch_bits):
        return self.REGISTERS[arch_bits]
    
    def create_legend(self):
        template = "Legend: | {stack} | {heap} | {code} | {instruction} | {symbol} |"
        stack = to_color("Stack", Color.STACK)
        heap = to_color("Heap", Color.HEAP)
        code = to_color("Code", Color.CODE)
        instruction = to_color("Instruction", Color.GREEN)
        symbol = to_color("Symbolic", Color.SYMBOLIC)
        template = template.format(stack=stack, heap=heap, code=code, instruction=instruction, symbol=symbol)
        return template 

    # (value, (value, ..))
    # TODO stack endness, fix
    def format_value(self, value):
        mem_location = self.check_ref(value)
        if not mem_location:
            return (self.value_to_pstr(value, mem_location), None)
        
        deref_value = self.deref(value, mem_location)
        value = self.value_to_pstr(value, mem_location)
        if deref_value != None:
            deref_value = self.format_value(deref_value)

        return (value, deref_value)
    
    def get_eval_sval(self, svalue):
        try:
            with time_limit(3):
                return hex(self.state.solver.eval(svalue))
        except Exception:
            return "eval timeout"
    def get_symbol_str(self, svalue):
        try:
            with time_limit(1):
                sym_string = str(svalue)
                if len(sym_string) > 40:
                    return sym_string[:40] + "..."
                return sym_string
        except Exception:
            return "To long to display"
    
    def sybmolic_var_representation(self, svalue):
        template = "<{value}> ({symbol_name})"
        value = self.get_eval_sval(svalue)
        symbol_name = self.get_symbol_str(svalue)
    
        return template.format(value=value, symbol_name=symbol_name)

    def value_to_pstr(self, value, mem_location):
        if value == None:
            return value
        if value.uninitialized:
            return to_color(str(value), Color.GRAY)
        elif value.symbolic:
            variable_name = self.sybmolic_var_representation(value)
            return to_color(str(variable_name), Color.SYMBOLIC)
        
        value = hex(value.concrete_value)
        if mem_location == MemLocation.STACK:
            return to_color(str(value), Color.STACK)
        elif mem_location == MemLocation.HEAP:
            return to_color(str(value), Color.HEAP)
        elif mem_location == MemLocation.CODE:
            return to_color(str(value), Color.CODE)
        else:
            return value

    def pprint(self):
        assert self.state is not None
        regs = {}
        for reg in self.REGISTERS[self.bits]:
            value = self.state.registers.load(reg)
            regs[reg] = self.format_value(value)
        
        pstr_regs = self.pprint_registers(regs)
        stack_objs = {}
        for i in range(8):
            offset = int(self.bits/8 * i)
            addr = self.stack_end + offset
            stack_objs[offset] = (addr , self.format_value(self.state.stack_read(offset, int(self.bits/8))))
        pstr_stack = self.pprint_stack(stack_objs)
        pstr_inst = self.pprint_instructions()
        legend = self.create_legend()
        return "\n" + legend + "\n" + pstr_regs + "\n" + pstr_stack + "\n" + pstr_inst

    def deref(self, value, mem_location):
        assert self.state is not None
        if mem_location == MemLocation.STACK or mem_location == MemLocation.HEAP:
            return self.state.memory.load(value, int(self.bits/8), endness=self.state.arch.memory_endness)

    # Check if the value is a reference to a value
    def check_ref(self, value):
        if value == None:
            return None
        elif value.symbolic:
            return None
        
        if self.stack_begin >= value.concrete_value >= self.stack_end:
            return MemLocation.STACK
        elif self.heap_start <= value.concrete_value <= self.heap_end:
            return MemLocation.HEAP
        elif self.code_start <= value.concrete_value <= self.code_end:
            return MemLocation.CODE
        else:
            return None
    
    def value_to_str(self, value, res_str="") -> str:
        item, ref = value
        if ref == None:
            return res_str + str(item)
        
        if isinstance(ref, tuple):
            res_str += str(item) + " --> "
            return self.value_to_str(ref, res_str)
        else:
            return res_str + str(item) + " --> " + str(ref)

        

    def pprint_registers(self, registers):
        register_str = "[%s]" % to_color(" Registers ", Color.CYAN).center(78, '-')
        register_str += "\n"
        for register in registers:
            value = registers[register]
            register_str += to_color(f"{register}: ", Color.BLUE)
            register_str += self.value_to_str(value) + "\n"

        return register_str


    # Check if value is 
    def pprint_stack(self, stack_objs):
        assert self.state is not None
        stack_str = "[%s]" % to_color(" Stack ", Color.CYAN).center(78, '-')
        stack_str += "\n"
        for offset in stack_objs:
            stack_str += to_color(f"{offset:04}| ", Color.BLUE)
            addr, value = stack_objs[offset]
            stack_str += f"{to_color(hex(addr), Color.STACK)} --> {self.value_to_str(value)}"
            if addr == self.bp:
                stack_str += '  <- bp'
            elif addr == self.stack_end:
                stack_str += '  <- sp'
            stack_str += "\n"
        
        return stack_str
    

    def pprint_instructions(self):
        assert self.state is not None
        instuction_str = "[%s]\n" % to_color(" Basic Block ", Color.CYAN).center(78, '-')
        try:
            block = self.state.block()
            basic_block_ins = block.disassembly.insns
        except Exception as e:
            return f"Error getting instructions: {e}"
        
        if block.size > 10:
            basic_block_ins = basic_block_ins[:10]
        
        for ins in basic_block_ins:
            instuction_str += to_color(f"{hex(ins.address)}: ", Color.CODE) + to_color(f"{ins.mnemonic}", Color.GREEN) + "\t" + f"{ins.op_str}"
            instuction_str += "\n" 

        if block.size > 10:
            instuction_str += to_color(f"...\n", Color.GRAY)
        
        return instuction_str



from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection


import pytest


import os
from unittest.mock import Mock


class TestDebugBreakpointCommands:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)

    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        assert dbg.handle("load example")
        return dbg
    
    
    def test_add_breakpoint(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert "Address 0x400566 added to breakpoints." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x400566.")
    
    
    def test_add_breakpoint2(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400560")
        assert "Address 0x400560 added to breakpoints." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x400560.")
    
    
    def test_add_breakpoint_at_line(self, dbg, conn):
        assert dbg.handle("add_breakpoint_at_line '/dangr/tests/example.c' 5")
        conn.send_info.assert_called_with("Address 0x400560 added to breakpoints.")
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Source Filter: /dangr/tests/example.c:5 (0x400560).")
    
    
    def test_remove_breakpoint(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert "Address 0x400566 added to breakpoints." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("remove_breakpoint 0x400566")
        assert "Address 0x400566 removed from breakpoints." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")

    
    def test_remove_non_breakpoint(self, dbg, conn):
        assert dbg.handle("remove_breakpoint 0x500")
        assert "Breakpoint at address 0x500 not found."  == str(conn.send_error.call_args[0][0])
    
    
    def test_enable_breakpoint(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("enable_breakpoint 0")
        assert "Breakpoint filter enabled." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x400566.")
    
    
    def test_enable_non_breakpoint(self, dbg, conn):
        assert dbg.handle("enable_breakpoint 0", False)
        assert "Index 0 out of range." == str(conn.send_error.call_args[0][0])

    
    def test_disable_breakpoint(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("disable_breakpoint 0")
        assert "Breakpoint filter disabled." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")

    
    def test_disable_non_breakpoint(self, dbg, conn):
        assert dbg.handle("disable_breakpoint 0", False)
        assert "Index 0 out of range." == str(conn.send_error.call_args[0][0])

    
    def test_list_breakpoints(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("list_breakpoints")
        assert "Breakpoint(s): [0] Address Filter: 0x400566" == str(conn.send_result.call_args[0][0])
       
    
    def test_list_no_breakpoints(self, dbg, conn):
        assert dbg.handle("list_breakpoints")
        assert "No breakpoints found." == str(conn.send_info.call_args[0][0])

    
    def test_clear_breakpoints(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400566")
        assert dbg.handle("clear_breakpoints")
        assert "All breakpoints cleared." == str(conn.send_info.call_args[0][0])
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Terminated.")
    
    


#generate documentation for the project:

from ..cli.cli_connection import CliConnection
from ..cli.command_line_debugger import CommandLineDebugger
from ..utils import utils

# requires installation of dAngr before running

# generate documentation for dAngr Commands
def document_gen():
    # Create md file with documentation
    with open("./docs/documentation.md", "w") as f:
        f.write("# dAngr Documentation\n\n")
        f.write("This documentation contains all the commands available in the dAngr debugger.\n\n")
        f.write('''
## General dAngr API
### Commands:
    * Commands are case-sensitive.
    * Commands can be called with or without the prefix.
    * Arguments are space separated.
### Variables and objects:
    * there are 4 types in dAngr:
        * &vars: variables that are stored in the debugger.
        * &sym: symbolic objects.
        * &mem: memory objects.
        * &reg: register objects.
    * variables can be created using assignment operator '='.
### Custom commands:
    * Custom commands can be created using the `def` keyword. Arguments specification is passed between brackets and comma separated. Followed by the body.
    * support for `if ...: body [else: body]`. 
    * support for `while ...: bod`y.
    * support for `for ...: body`.

The body is indented on newlines
    Example:
    ```
    def my_command(arg1, arg2):
        print(arg1)
                
''')

        f.write("## Built-in Debugger Commands\n")

        dbg = CommandLineDebugger(CliConnection())
        data, _ = dbg.list_commands_data(True)

        for row in data:
            for d in row:
                d = d.replace('<command>', '**').replace('</command>', '**')
                d = utils.remove_xml_tags(d).replace('\t', '&nbsp;'*4)
                f.write(f"{d}")
            f.write("\n\n")
        
        f.write("\n")
        f.write('''
## Python API
* Execute python commands in the debugger by preceding with `!` (exclamation mark). For example, to print the value of a variable `x`, use `!print(x)`.
 Variables are converted into strings before being passed to the python interpreter.
* Variables can be passed using the long variable name (including the prefix &var, &sym, etc.).
* Python calls may include dAngr or shell commands by using &(...), $(...) respectively.

## Bash API
* Execute shell commands in the debugger by preceding with `%` (percentage sign). For example, to list the files in the current directory, use `%ls`.
* include dAngr or Python commands by using &(...), !(...) respectively.


''')
    print("Documentation generated successfully.")

document_gen()